#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
One-shot wrapper, where EMLM is replaced with a DNN network

@author: Joakim Linja

"""

import numpy as np
from typing import List
import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from sklearn import datasets
from ML_functions import preprocess
from copy import deepcopy 


def compileDNN(inputShape: int, layers: List[int],useBias=False):
    """
    """
    
    model = Sequential()
    model.add(tf.keras.Input(shape=inputShape))
    for layer in layers[:-1]:
        model.add(Dense(layer, activation='tanh',use_bias=False))
    model.add(Dense(layers[-1], activation='linear',use_bias=False))
    model.compile(optimizer='adam',
                  loss=tf.keras.metrics.mean_squared_error,
                  metrics=[tf.keras.metrics.RootMeanSquaredError()])
    return model


def trainDNN(model, X, Y):
    """
    """
   
    model.fit(X, Y, verbose=0,
              use_multiprocessing=True,
              )
    return model


def DNN_MAS_N(layers, X, Y, weights, biases, model):
    """
    """
    
    if len(layers)==2:
        return DNN_MAS_2(X, Y, weights, biases, model)[0]
    elif len(layers)==4:
        return DNN_MAS_4(X, Y, weights, biases, model)[0]
    

def DNN_MAS_N2(layers, X, Y, verbose=0, verification=True):
    """
    """
    
    (xI,xJ) = np.shape(X)
    model = compileDNN(xJ,layers)
    model = trainDNN(model,X,Y)
    (weights,biases) = layersFromModel(model)
    
    if len(layers)==2:
        res = DNN_MAS_2(X, Y, weights, biases, model, verbose=verbose, verification=verification)
    elif len(layers)==4:
        res = DNN_MAS_4(X, Y, weights, biases, model, verbose=verbose, verification=verification)
    (r1,r2) = res
    return r1[0],r2


def DNN_MAS_4(X, Y, weights, biases, model, verbose=0, verification=False):
    """
    """
    
    print('\nDNN-MAS-4')
    (N,n0) = np.shape(X)
    Yshape = np.shape(Y)
    if len(Yshape)==1:
        n2 = Yshape[0]
    else:
        (_,n2) = Yshape
    
    W1 = (weights[0]).T
    W2 = (weights[1]).T
    W3 = (weights[2]).T
    W4 = (weights[3]).T
    #b1 = biases[0]
    #b2 = biases[1]
    
    #p = np.transpose(np.hstack((np.ones((N,1)),X)))
    #W1Xpart = W1 @ X.T
    WF = W1 @ X.T
    F = 2/(1+np.exp(-2*WF))-1
    D1 = 1-np.power(F,2)
    
    WF = W2 @ F
    F = 2/(1+np.exp(-2*WF))-1
    D2 = 1-np.power(F,2)
    
    WF = W3 @ F
    F = 2/(1+np.exp(-2*WF))-1
    D3 = 1-np.power(F,2)
    
    #WF = W4 @ F
    #F = 2/(1+np.exp(-2*WF))-1
    #D4 = 1-np.power(F,2)

    
    cof = 1/N
    DYX = np.zeros((n2,n0)) # MAS values
    verif = -1*np.ones((N,n0))
    for i in range(N):
        #D1 = np.array(list(map(lambda x: x*(1-x),F1[:,i]))) #F1[:,i]*(1-F1[:,i])
        D1i = D1[:,i]
        D2i = D2[:,i]
        D3i = D3[:,i]
        #D4i = D4[:,i]
        #DYX += cof*np.abs( W2[:,1:n1+1] @ np.diag(D1i) @ W1[:,1:n0+1] )
        
        #DYX = DYX + cof*abs(W4(:,2:n3+1)*diag(D3)*W3(:,2:n2+1)*diag(D2)*W2(:,2:n1+1)*diag(D3)**W1(:,2:n0+1));
        #DYX += cof*np.abs( W2 @ np.diag(D1i) @ W1 )
        abspart = W4 @ np.diag(D3i) @ W3 @ np.diag(D2i) @ W2 @ np.diag(D1i) @ W1
        
        if verification:
            Xi = X[i,:].reshape(1,-1)
            abspartj = np.zeros(np.shape(abspart))
            diff = 1e-3
            for j in range(n0):
                Xp = deepcopy(Xi)
                Xm = deepcopy(Xi)
                Xp[0,j] = Xi[0,j]+diff
                Xm[0,j] = Xi[0,j]-diff
                
                rp = model.predict(Xp,verbose=0)
                rm = model.predict(Xm,verbose=0)
                
                abspartj[0,j] = (rp-rm)/(2*diff)
                
            temp = abspart[0,:]-abspartj[0,:]
            verif[i,:] = temp
        
        DYX += cof*np.abs(abspart)
    
    return DYX, verif


def DNN_MAS_2(X, Y, weights, biases, model, verbose=0, verification=False):
    """
    """
    print('\nDNN-MAS-2')
    (N,n0) = np.shape(X)
    Yshape = np.shape(Y)
    if len(Yshape)==1:
        n2 = Yshape[0]
    else:
        (_,n2) = Yshape
    
    W1 = (weights[0]).T
    W2 = (weights[1]).T
    #b1 = biases[0]
    #b2 = biases[1]
    
    #p = np.transpose(np.hstack((np.ones((N,1)),X)))
    #W1Xpart = W1 @ X.T
    W1X = W1 @ X.T
    
    #W1X = np.hstack((b1.reshape(-1,1),W1Xpart))
    #F1 = 1/(1+np.exp(-W1X)) # sigmoid
    F1 = 2/(1+np.exp(-2*W1X))-1# tanh
    
    #H = np.vstack((np.ones(1,N),F1))
    
    D1 = 1-np.power(F1,2)
    
    cof = 1/N
    DYX = np.zeros((n2,n0)) # MAS values
    
    verif = -1*np.ones((N,n0))
    for i in range(N):
        #D1 = np.array(list(map(lambda x: x*(1-x),F1[:,i]))) #F1[:,i]*(1-F1[:,i])
        D1i = D1[:,i]
        #DYX += cof*np.abs( W2[:,1:n1+1] @ np.diag(D1i) @ W1[:,1:n0+1] )
        abspart = W2 @ np.diag(D1i) @ W1
        
        if verification:
            Xi = X[i,:].reshape(1,-1)
            abspartj = np.zeros(np.shape(abspart))
            diff = 1e-3
            for j in range(n0):
                Xp = deepcopy(Xi)
                Xm = deepcopy(Xi)
                Xp[0,j] = Xi[0,j]+diff
                Xm[0,j] = Xi[0,j]-diff
                
                rp = model.predict(Xp,verbose=0)
                rm = model.predict(Xm,verbose=0)
                
                abspartj[0,j] = (rp-rm)/(2*diff)
                
            temp = abspart[0,:]-abspartj[0,:]
            verif[i,:] = temp
        
        DYX += cof*np.abs( abspart )
    
    return DYX, verif


def irisflowerDataset():
    dat = datasets.load_iris()
    X = dat.data
    Y = dat.target
    Y = Y.reshape(-1,1)
    
    X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask = \
        preprocess(X,Y, unmodifiedOriginals=True,
                   Xinterval=[-1,1], Yinterval=[-1,1],constThreshold=1e-15)
    
    mask = np.ones(len(nY),dtype=bool)
    mask = np.concatenate((np.arange(40),np.arange(50,90),np.arange(100,140)))
    cmask = np.concatenate((np.arange(40,50),np.arange(90,100),np.arange(140,149)))
    
    trnX = nX[mask,:]
    trnY = nY[mask]
    tstX = nX[cmask,:]
    tstY = nY[cmask]
    
    return trnX, trnY, tstX, tstY


def layersFromModel(mdl):
    """
    """
    ws = []
    bs = []
    layers = mdl.layers
    i=0
    for layer in layers: 
        weights = layer.get_weights()
        ws.append(weights[0])

        i+=1
    return ws, bs


if __name__ == "__main__":
    
    # Using Iris dataset as a commonly available reference, even though here 
    # it is used as a regression dataset
    trnX, trnY, tstX, tstY = irisflowerDataset()
    (N,m) = np.shape(trnX)
    
    featn = m
    #layers = [int(np.ceil(featn/2)), int(np.ceil(featn/4)), int(np.ceil(featn/8)), 1]
    layers = [int(np.ceil(featn/2)), 1]
    
    DYX, verif = DNN_MAS_N2(layers, trnX, trnY)
    print(DYX)
    print(verif)
    
    print('\nEnd of Line')