#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A collection of functions used in machine learning

@author: Joakim Linja
"""
import numpy as np
from scipy.spatial.distance import cdist
import random
from copy import deepcopy
from matlabLike import pdist2
from sklearn.utils.extmath import randomized_svd
from pandas import DataFrame
from scaling import MinMax
import scipy as sp
from itertools import cycle
#from pairwiseDistanceCalculation import pdc


def RMSE(Yreal: np.ndarray, Ypred: np.ndarray):
    (a,b) = Yreal.shape
    (c,d) = Ypred.shape
    assert a==c
    assert b==d
    return np.sqrt(np.mean(np.square(Yreal-Ypred)))


def MRSE(Yreal: np.ndarray, Ypred: np.ndarray):
    (a,b) = Yreal.shape
    (c,d) = Ypred.shape
    assert a==c
    assert b==d
    return np.mean(np.sqrt(np.square(Yreal-Ypred)))


def MAE(Yreal: np.ndarray, Ypred: np.ndarray):
    (a,b) = Yreal.shape
    (c,d) = Ypred.shape
    assert a==c
    assert b==d
    return np.mean(np.abs(Yreal-Ypred))


def removeDuplicates(X: np.ndarray) -> np.ndarray:
    """
    Uses a ready made function from Pandas
    
    Parameters
    ----------
    *X* : numpy.ndarray
        Input data matrix
        
    Returns
    -------
    *data* : numpy.ndarray
        *X*, but without duplicate observations (rows)
    """
    
    df = DataFrame(X)
    df.drop_duplicates(subset=None, keep='first', inplace=True)
    data = df.to_numpy(dtype=np.float, copy=False)
    return data


def preprocess(X, Y, printT=False, unmodifiedOriginals=False,
               Xinterval=[0,1], Yinterval=[0,1], constThreshold=1e-15):
    """
    Standard preprocessing to a dataset.
    Removal of duplicates and constants
    Scaling of individual features to [0,1] (default) scale.

    Returns X, Y, normalized X and y along with the scaler-objects.
    
    Parameters
    ----------
    
    *X* : numpy.ndarray
        Input data observation matrix
        
    *Y* : numpy.ndarray
        Input data targets
        
    *printT* : boolean, optional
        If true, prints out the number of removed duplicates and constant features
        
        Default False
        
    *unmodifiedOriginals* : bool, optional
        Defines whether the returned *X* and *Y* will have constants removed
        Default: False
        
    *Xinterval* : tuple, optional
        Defines the range of values the features in *X* are scaled to
        
        Default: [0,1]
        
    *Yinterval* : tuple, optional
        Defines the range of values the targets in *Y* are scaled to
        
        Default: [0,1]
        
    *constThreshold* : float, optional
        Defines the variance threshold for what is considered a constant feature
        
        Default: 1e-15
    
    Returns
    -------
    
    *X* : numpy.ndarray
        The input data observation matrix
        
    *Y* : numpy.ndarray
        The input data targets
        
    *nX* : numpy.ndarray
        The preprocessed version of *X*
        
    *nY* : numpy.ndarray
        The preprocessed version of *Y*
        
    *Xscaler* : Scaler-object
        The Scaler-object used to create *nX*
        
    *Yscaler* : Scaler-object
        The Scaler-object used to create *nY*
        
    *XconstFiltMask* : numpy.ndarray
        Boolean array used to filter out constant features from *X*
        
    *YconstFiltMask* : numpy.ndarrays
        Boolean array used to filter out constant targets from *Y*
        Is ineffective in single target situations. 
    """

    (NX, wX) = np.shape(X)
    (NY, wY) = np.shape(Y)

    assert wX != 0
    assert wY != 0
    singleOutput = wY==1
    XY = removeDuplicates(np.hstack((X,Y)))
    (NXY,_) = np.shape(XY)
    X = XY[:,:wX]
    Y = XY[:,wX:]
    assert np.shape(X)[1] == wX
    assert np.shape(Y)[1] == wY

    X, XconstFiltMask = removeConstants(
            X,absolute=False, threshold=constThreshold,
            unmodifiedOriginals=unmodifiedOriginals)

    if not singleOutput:
        Y, YconstFiltMask = removeConstants(
            Y,absolute=False, threshold=constThreshold,
            unmodifiedOriginals=unmodifiedOriginals)
    else:
        YconstFiltMask = np.array([True])

    Xscaler = MinMax(interval=Xinterval)
    Yscaler = MinMax(interval=Yinterval)

    if unmodifiedOriginals:
        nX = Xscaler.Scaling(X[:,XconstFiltMask])
        nY = Yscaler.Scaling(Y[:,YconstFiltMask])
    else:
        nX = Xscaler.Scaling(X)
        nY = Yscaler.Scaling(Y)

    if printT:
        print('Removed',NX-NXY,'duplicate observations')
        print('Removed',wX-np.shape(X)[1],'constant input features')

    return X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask


def _rankKrandomSVD(A: np.matrix, k: int):
    """
    https://amath.colorado.edu/faculty/martinss/Talks/2010_mmds_martinsson.pdf
    """
    _, n = A.shape
    # 1. O: n x k
    O = np.random.normal(size=(n,k))
    # 2. Y = A @ O, Y: m x k
    #Y = A @ O
    # 3. Q: m x k, Y = Q @ R
    Q, _ = np.linalg.qr(A @ O, mode='reduced')#Y,mode='reduced')
    # 4. B = Q @ A, B: k x n
    #B = Q @ A
    # 5. B = U1 @ S @ Vh
    U1, S, Vh = np.linalg.svd(Q.T @ A, full_matrices=False)#B,full_matrices=False)
    # 6. U = Q @ U1
    return Q @ U1, S, Vh


def sk_rand_svd(p: float, I, A, b, rndState=0):
    """
    p = percentage of singular values to extract
        
    I =  number of iterations
    
    A = matrix
    
    b = vector
    
    (U, S, V.T) ~= Dx
    B = V @ S^-1 @ U @ Dy
    
    
    Random state defaults to 0. Official default behaviour is "None"
    """
    U,S,Vh = randomized_svd(A,
                n_components=(int(A.shape[0]*p)),n_iter=I,random_state=rndState)
    return np.multiply(Vh.T, 1/S) @ U.T @ b


def _rSVDstngsExtractor(mtdstr):
    parts = mtdstr.split('-')
    if len(parts)==3:
        try:
            I = int(parts[2])
        except ValueError:
            I = parts[2]
        p = float(parts[1])
    else:
        # Only method was given, defaulting to: 
        p = 1
        I = 'auto'
    return (p,I)


def solve_Axb(A,b,method='svd',blockN=-1):
    """
    Alias for various Ax=b solvers. With rSVD, method string is to be
    "rSVD-p-I" where p (float) is remaining rank percentage and I is iteration count.
    If string is "rSVD-p", I is assumed as 1.
    """
    if method=='lstsq':
        return np.linalg.lstsq(A,b,rcond=None)[0]
    if method=='lstsqF':
        return np.linalg.lstsq(A,b,rcond=None)
    if method=='slstsq':
        return sp.linalg.lstsq(A,b)[0]
    if method=='svd':
        U, S, Vh = np.linalg.svd(A, full_matrices=False)
        if blockN==-1:
            return np.multiply(Vh.T, 1/S) @ U.T @ b
        else:
            return blockMatMul(np.multiply(Vh.T, 1/S),blockMatMul(U.T,b,blockN=blockN),blockN=blockN)
    if 'rSVD' in method:
        (p,I) = _rSVDstngsExtractor(method)
        return sk_rand_svd(p,I,A,b)

    raise NotImplementedError('Faulty method selector {0:s} given'.format(
            method))


def blockMatMul(A,B,blockN=2):
    """
    Divide A and B into N block matrices and perform A @ B blockwise
    """
    
    (nai,naj) = np.shape(A)
    (nbi,nbj) = np.shape(B)
    assert naj==nbi
    # horizontal split for A
    p1 = int(np.floor(nai/blockN))
    p2 = nai%blockN
    sizesA = [p1 for i in range(blockN)]
    sizesA[-1] += p2
    
    # vertical split for B
    p1 = int(np.floor(nbj/blockN))
    p2 = nbj%blockN
    sizesB = [p1 for i in range(blockN)]
    sizesB[-1] += p2
    
    res = np.zeros((nai,nbj))
    for i in range(blockN):
        for j in range(blockN):
            sa0 = np.sum(sizesA[:i],dtype=int)
            sa1 = sa0+sizesA[i]
            sb0 = np.sum(sizesB[:j],dtype=int)
            sb1 = sb0+sizesB[j]
            bA = A[sa0:sa1,:]
            bB = B[:,sb0:sb1]
            res[sa0:sa1,sb0:sb1] = bA @ bB
            
    return res
    

def dobscv_MC(Classes, Data, nfolds, deterministic=False):
    """
    **Multi-class DOB-SCV implementation**

    Original doctext:
    % Creation of partition of data using DOB-SCV approach
    % (c) TommiK
    %   This version storing Labels into two separate DSs due to Oct 3, 2017

    """
    if deterministic: # new addition
        # Take the one closest to the mean and use it as a starting point
        def firstFun(dat,_):
            mean = np.mean(dat,axis=0)
            norm = np.linalg.norm(dat-mean,axis=1)
            srtI = np.argmin(norm)
            return dat[srtI]
    else:
        def firstFun(dat,num):
            return dat[np.random.choice(num),:]

    #N = length(Classes);
    N = len(Classes)
    #ncls = max(Classes);
    ncls = np.max(Classes)

    #FoldLabels = zeros(N,1);
    FoldLabels = np.zeros((N,1),dtype=np.int)-1000 # -maxint so that if there is any
    # negative values in the end, something is wrong

    #FoldLbs = [1:nfolds]';
    FoldLbs = np.arange(nfolds).reshape(-1,1)
    #for c=1:ncls
    for c in range(ncls+1):
        # Ic = find(Classes == c);
        Ic = np.argwhere(Classes==c).reshape(1,-1)[0]
        # numobs = length(Ic);
        numobs = len(Ic)
        # CData = Data(Ic,:);
        CData = Data[Ic,:]
        # while (numobs >= nfolds),
        while numobs >= nfolds:
            # e0 = CData(randperm(numobs,1),:);
            #e0 = CData[np.random.choice(numobs),:]
            e0 = firstFun(CData,numobs)
            # [~,I] = pdist2(CData,e0,'euclidean','Smallest',nfolds);
            (_, I) = pdist2(CData, e0.reshape(1,-1), sort='Smallest', n=nfolds)
            # J = Ic(I);
            J = Ic[I]
            # FoldLabels(J) = FoldLbs;
            FoldLabels[J[:,0],0] = FoldLbs[:,0]
            # Ic = setdiff(Ic,J);
            Ic = np.setdiff1d(Ic, J)
            # numobs = length(Ic);
            numobs = len(Ic)
            # CData = Data(Ic,:);
            CData = Data[Ic,:]

        # FoldLabels(Ic) = [1:numobs]';
        FoldLabels[Ic] = np.arange(0,numobs).reshape(-1,1)

    #%Storing labels separately into training and test sets
    #CVTrain = ones(nfolds,N); CVTest = zeros(nfolds,N);
    CVTrain = np.ones((nfolds,N),dtype=np.int)
    CVTest = np.zeros((nfolds,N),dtype=np.int)
    #for i=1:nfolds
    for i in range(nfolds):
        # I = find(FoldLabels == i);
        I = np.argwhere(FoldLabels==i)[:,0]
        # CVTest(i,I) = 1;
        CVTest[i,I] = 1
        # CVTrain(i,I) = 0;
        CVTrain[i,I] = 0

    #%Changing type to save space and make ready for subset/sample selection
    #CVTrain = logical(CVTrain); CVTest = logical(CVTest);
    CVTrain = CVTrain.astype(np.bool)
    CVTest = CVTest.astype(np.bool)

    assert np.sum(FoldLabels<0)==0
    return CVTrain,CVTest,FoldLabels


def MinSorter(X,k):
    """
    Assumes that shape(X) == (N,1)

    X = array to be sorted and divided

    k = number of total folds

    Returns

    (valInds,valX), (trnInds,trnX)

    valInds

    valX

    trnInds

    trnX

    """
    X = deepcopy(X)
    X1 = X.reshape((-1,))
    N, w = np.shape(X)
    assert w==1
    selInds = np.argsort(X1)

    inds = []
    for i in range(k):
        inds.append(selInds[(k-i-1)::k])
    valInds = inds[0]
    trnInds = np.array([],dtype=np.int)
    for i in range(1,k):
        trnInds = np.hstack((trnInds,inds[i]))

    valX = X[valInds,:]
    trnX = X[trnInds,:]

    return (valInds,valX), (trnInds,trnX)


def MinSorter2(X,k,sel=-2):
    """
    """
    X = deepcopy(X)
    X1 = X.reshape((-1,))
    N, w = np.shape(X)
    assert w==1
    assert k>2
    selInds = np.argsort(X1)

    inds = []
    for i in range(k):
        inds.append(selInds[i::k])

    valInds = inds[sel]
    indarr = -1*(1+np.arange(k))
    trnInds = np.array([],dtype=np.int)
    for i in np.flip(indarr):
        if i==sel:
            continue
        trnInds = np.hstack((trnInds,inds[i]))

    valX = X[valInds,:]
    trnX = X[trnInds,:]
    return (valInds,valX), (trnInds,trnX)



def FoldSortMin(X,k):
    """
    """
    #print('FoldSortMin')
    # 19//5 = 3
    # 19%5 = 4
    #print('shape(X):',np.shape(X))

    X = deepcopy(X)
    X1 = deepcopy(X)
    X1 = X1.reshape((-1,))
    N, w = np.shape(X)

    siX = np.argsort(X1)

    # k folds are required, get remainder
    rem = N%k
    fs = N//k # fold size

    #print('shape(X):',np.shape(X))
    #print('shape(X1):',np.shape(X1))
    #print('shape(siX):',np.shape(siX))
    #print('rem:',rem)
    #print('fs:',fs)

    if False:#rem==0:
        # Initialize Folds and rInds
        Folds = np.zeros((k,fs))
        rInds = np.zeros((k,fs),dtype=np.int)
        rInds[:,:] = siX.reshape((k,fs))
        for i in range(k):
            Folds[:,i] = X[rInds[i,:],0]

        resF = []
        resI = []
        for i in range(k):
            resF.append(Folds[i,:].reshape(-1,1))
            resI.append(rInds[i,:].reshape(-1,))
        Folds = resF
        rInds = resI

    else:
        Folds = []
        rInds = []
        j=rem
        for i in range(k):
            if j >0:
                Folds.append(np.zeros((fs+1,1)))
                rInds.append(np.zeros((fs+1,),dtype=np.int))
                j-=1
            else:
                Folds.append(np.zeros((fs,1)))
                rInds.append(np.zeros((fs,),dtype=np.int))


        i = cycle(range(k))
        j = cycle(range(fs))
        j0 = 0
        i0 = 0
        for e in siX:
            i0 = next(i)
            rInds[i0][j0] = e
            if i0==k-1:
                j0 = 1+next(j)

        for i in range(k):
            rI = rInds[i]
            for j in range(len(rI)):
                Folds[i][j] = X[rI[j],0]

    return Folds, rInds



def dobscv_SC(X, k, deterministic=False):
    """
    **Single-class DOB-SCV implementation.**

    **Input:**

        - *X*: data matrix of all observations to be used
        - *k*: into how many "uniform" groups is the data to be split into

    **Output:**

        - *Folds*: *X* split into *k* groups
        - *rInds*: Original indices of *X* in the same shape and order as in
          *Folds*

    **Sources:**

    Page 6 of http://dx.doi.org/10.1016/j.ins.2013.09.038
    "On the importance of the validation technique for classification
    with imbalanced datasets: Addressing covariate shift when data
    is skewed"
    Algorithm 1:
    ```
    {
    for each class c_j \in C do
        while count (c_j) > 0 do
            e_0 <- randomly select an example of class c_j from D
            e_i <- i:th closest example to e_0 of class c_j from D (i=1,...,k-1)
            F_i <- F_i \\union e_i (i=0,...,k-1)
            D <- D\e_i (i=0,...,k-1) # Remove e_i from D
        end while
    end for
    }
    ```

    Page 30 of Joonas-Hämäläinen-Väitös-978-951-39-7621-7_vaitos14122018.pdf
    Algorithm 3:

    Input: Dataset X and the number of folds k.
    Output: Folds F_1 , F_2 , ..., F_k .
    1: while X has more than k points do
    2:  e_1 <- select a point uniformly at random from X.
    3:  { e_2 , ..., e_k } <- find k−1 nearest neighbors of e_1 from X.
    4:  F_i <- F_i \\union e_i for each i = { 1, 2, ..., k } .
    5:  X <- X \{ e_1 , ..., e_k } .
    6: Set rest of the points from X into different folds.

    Implementation: @Joakim Linja
    """
    if deterministic: # new addition
        # Take the one closest to the mean and use it as a starting point
        def firstFun(_,datX):
            if len(np.shape(datX))==1:
                mean = np.mean(datX)
                norm = np.linalg.norm(datX-mean)
            else:
                mean = np.mean(datX,axis=0)
                norm = np.linalg.norm(datX-mean,axis=1)
            srtI = np.argmin(norm)
            return srtI
    else:
        def firstFun(datI,_):
            return np.random.choice(datI)

    if deterministic:
        print('Deterministic DOB-SCV')
    else:
        print('Random DOB-SCV')

    X = deepcopy(X)
    N, w = np.shape(X)
    Ns = int(np.floor(N/k))
    kc = 0 # running k-index
    

    # Initialize Folds and rInds almost completely
    # If X%k==0 then no leftovers, otherwise leftovers need additional treatment
    Folds = np.zeros((k,Ns,w))
    rInds = np.zeros((k,Ns),dtype=np.int)

    # Initialize original index list
    Inds = np.arange(0,N)
    # Loop until there are <=k observations left
    while len(Inds) >= k:

        # Current indices of observations in X
        XindList = np.arange(0,len(X))
        
        # Choose one observation randomly/deterministically
        rndChI = firstFun(XindList,X)
        # Find the indices of the k-nearest neighbours of the random observation
        #en = kNN(X,X[rndChI,:],k)
        en = kNN(X,X[rndChI,:],k)
        #print('en:',en)
        # *en* now contains indices of the observations that are to be
        # added to Folds
        #assert len(en)==k

        # Union operations
        Folds[:,kc,:] = X[en,:]
        rInds[:,kc] = Inds[en]
        kc +=1
        # Separation operations
        #  Remove added indices from the remaining indices to create a filter
        sen = np.array(list(filter(lambda y: y not in en, XindList)))
        if len(sen)<1:
            X = []
            Inds = []
        else:
            X = X[sen,:]
            Inds = Inds[sen]

    # From here on, the topmost level of Folds and rInds cannot be an array,
    # as they are not equisized
    Folds = list(Folds)
    rInds = list(rInds)

    # Add leftovers to random folds
    linds = np.arange(0,k)
    random.shuffle(linds)
    for i in range(len(X)):
        Folds[linds[i]] = np.vstack((Folds[linds[i]],X[i]))
        rInds[linds[i]] = np.append(rInds[linds[i]],Inds[i])

    return Folds, rInds


def removeConstants(X:np.ndarray, absolute=True,
                    threshold=1e-5, unmodifiedOriginals=False):
    """
    Removes constants and returns filtered with filter mask

    Returns (X, mask)
    """
    tX = type(X)
    if absolute:
        check = lambda y: np.var(y,axis=0)!=0
    else:
        check = lambda y: (
                np.abs(np.max(y,axis=0)-np.min(y,axis=0))>=threshold)

    if tX==np.mat:
        mask = np.array(check(X))[0]
        if not unmodifiedOriginals:
            X = (X.T[mask]).T
    elif tX==np.ndarray:
        mask = check(X)
        if not unmodifiedOriginals:
            X = np.transpose(np.transpose(X)[mask])
    else:
        raise NotImplementedError

    return X, mask



def kNN(X,x,k):
    """
    Input:

        *X* = Remaining observations, data
        *x* = randomly chosen observation from *X*
        *Inds* = original indices of the remaining observations in *X*
        *ind* = current index of *x* in *X*
        *k* = nearest neighbor count

    Output:
        *en* = list of *k* nearest neighbors indices with the index of *x*
    """
    # Calculate distances of *x* to all in *X* and convert the
    # result from 2D into 1D
    inds = np.arange(len(X))
    dists = cdist(X,[x]).flatten()
    #dists = pdc(X,[x]).flatten()

    # argsort forms a list of indices that would sort
    # the given *dists*. This list is then shortened to k-neares neighbours
    # and then used to filter out the ordered indices from *inds*
    # heapsort is speed 3, O(n*log(n)) with 0 workspace according to docs
    # Another options:
    #  kind 	  speed worst case 	work space 	stable
    #  ‘quicksort’ 	1 	O(n^2) 	     0 	        no
    #  ‘heapsort’ 	3 	O(n*log(n))  0 	        no
    #  ‘mergesort’ 	2 	O(n*log(n))  ~n/2 	    yes
    #  ‘timsort’ 	2 	O(n*log(n))  ~n/2 	    yes
    #
    return inds[(np.argsort(dists,kind='heapsort'))[0:k]]


def rsmaximin(X, n1):
    """
    Original doctext below:

    function [Inds,Lastd] = rsmaxmin(X,n1)
    %
    % Reimplementation of rsmaxmin reference point selection method from MLM manuscript and NeCo article
    %

    """
    #[N,n] = size(X);
    (N,n) = np.shape(X)
    #Inds = zeros(n1,1);
    Inds = np.zeros((n1,1),dtype=np.int)
    #[~,ind] = min(pdist2(X,mean(X)));
    tpd2 = pdist2(X,np.mean(X,axis=0).reshape(1,-1))
    ind = np.argmin(tpd2)

    indices = np.arange(N).reshape(-1,1)

    #Inds(1) = ind;
    Inds[0] = ind
    #%
    #Dist = inf(N,1); FillRest = 0;
    Dist = np.full((N,1),float("inf"))
    FillRest = 0
    #for i = 2:n1
    for i in range(1,n1):
        # Dist = min(Dist,pdist2(X,X(Inds(i-1),:)));
        #tx = X[Inds[i-1],:]
        tx = (X[Inds[i-1,0],:]).reshape(1,-1)
        tpd = pdist2(X, tx)
        Dist = np.min(np.hstack((Dist, tpd)),
                      axis=1).reshape(-1,1)


        # J = find(Dist>0);
        tJ = Dist > 0
        J = indices[tJ[:,0],:]
        # [Lastd,ind] = max(Dist(J));
        tDist = Dist[J[:,0]]
        Lastd = np.max(tDist)
        ind = np.argmax(tDist)

        # if (~isempty(J))
        if not len(J)==0:
            # Inds(i) = J(ind);
            Inds[i,0] = J[ind,0]
        # else
        else:
            # Lastd = 0;
            Lastd = 0
            # FillRest = 1;
            FillRest = 1
            # break
            print('BREAK')
            break

    #if (FillRest)
    if FillRest==1:
        # J = setdiff(1:N,Inds);
        J = np.setdiff1d(np.arange(N),Inds)
        # Inds(i:n1) = J(randperm(length(J),n1-i+1));
        Inds[i:n1] = J[np.random.permutation(len(J)),n1-i+1]

    return Inds, Lastd


def weightedEuclideanDistance(A, B, W):
    """
    $$d(A,B) = \sqrt{ \sum_i w_i (A_i - B_i)^2 }$$
    """
    nma = np.shape(A)
    nmb = np.shape(B)
    nmw = np.shape(W)
    if len(nma)==1:
        assert nma == nmb
        assert nma == nmw
    elif len(nma)==2:
        assert nma[0]==nmb[0]
        assert nma[0]==nmw[0]
        assert nma[1]==nmb[1]
        assert nma[1]==nmw[1]
    else:
        raise NotImplementedError
    return np.sqrt(np.sum(W*W*(A-B)**2))


def __test_weightedEuclideanDistance():
    a = np.random.rand(15)
    b = np.random.rand(15)
    w = np.random.rand(15)
    d1 = weightedEuclideanDistance(a,b,w)
    a = a.reshape(-1,1)
    b = b.reshape(-1,1)
    w = w.reshape(-1,1)
    d2 = weightedEuclideanDistance(a,b,w)
    a = a.reshape(1,-1)
    b = b.reshape(1,-1)
    w = w.reshape(1,-1)
    d3 = weightedEuclideanDistance(a,b,w)
    print(d1)
    print(d2)
    print(d3)


def __test_dobscv_SC():
    a = np.zeros((6,6),dtype=np.int)
    for i in range(6):
        a[i,:] = i
    a = a.reshape((-1,))
    random.shuffle(a)
    a = a.reshape((-1,1))
    (Folds, rInds) = dobscv_SC(a, 6)
    print(Folds)
    print(rInds)

def __test_FoldSortMin():
    """
    """
    # Test 1: 6x6
    from itertools import cycle
    a = np.zeros((6,6),dtype=np.int)
    for i in range(6):
        a[i,:] = i
    a = a.reshape((-1,))
    random.shuffle(a)
    a = a.reshape((-1,1))
    (Folds, rInds) = FoldSortMin(a, 6)
    check = np.arange(6)
    for fold in Folds:
        assert 6 == len(fold)
        for n in check:
            assert n in fold
    cri = set()
    for rInd in rInds:
        for rIn in rInd:
            cri.add(rIn)
    assert len(cri)==36
    print('Passed test 1: 6x6')

    # Test 2: 6 folds, 40 elements
    nums = cycle(np.arange(6))
    a = np.zeros(40)
    for j in range(len(a)-4):
        a[j] = next(nums)
    a[36:40] = np.arange(6,10)
    random.shuffle(a)
    a = a.reshape((-1,1))
    (Folds, rInds) = FoldSortMin(a, 6)

    # Check that each index is unique
    cri = set()
    for rInd in rInds:
        for rIn in rInd:
            cri.add(rIn)
    assert len(cri)==40

    # Check that each value is present the correct amount
    numsC = {0:0,1:0,2:0,3:0,4:0,5:0,6:0,7:0,8:0,9:0}
    for i in range(6):
        for j in range(len(rInds[i])):
            t = Folds[i][j,0]
            numsC[t]+=1
    for n in numsC:
        if n in range(6):
            assert 6 == numsC[n]
        else:
            assert 1 == numsC[n]

    # Check that each Fold has the correct number of values
    for i in range(6):
        t = Folds[i][:,0]
        s = set(t)
        if i<4:
            assert len(s) == 7
        else:
            assert len(s) == 6

    print('Passed test 2: 6 folds, 40 elements')

    # Test 3: 6x6, unique ends
    from itertools import cycle
    a = np.zeros((6,6),dtype=np.int)
    for i in range(5):
        a[i,:] = i
    b = 5
    for i in range(6):
        a[-1,i] = b
        b+=1
    #print(a)

    a = a.reshape((-1,))
    random.shuffle(a)
    a = a.reshape((-1,1))
    (Folds, rInds) = FoldSortMin(a, 6)
    #print(Folds)
    #print(rInds)
    cri = set()
    for rInd in rInds:
        for rIn in rInd:
            cri.add(rIn)
    assert len(cri)==36

    # Check that each value is present the correct amount
    numsC = {0:0,1:0,2:0,3:0,4:0,5:0,6:0,7:0,8:0,9:0,10:0}
    for i in range(6):
        for j in range(len(rInds[i])):
            t = Folds[i][j,0]
            numsC[t]+=1
    for n in numsC:
        if n in range(5):
            assert 6 == numsC[n]
        else:
            assert 1 == numsC[n]

    for i in range(6):
        t = set(Folds[i][:,0])
        assert len(t)==6
        assert 0 in t
        assert 1 in t
        assert 2 in t
        assert 3 in t
        assert 4 in t
        assert 5+i in t

    print('Passed test 3: 6x6+unique')


def __test_MinSorter():
    """
    """
    a = np.arange(10)
    b = np.arange(10)
    random.shuffle(b)

    #print('a:',a)

    (avi,avx), (ati,atx) = MinSorter(a.reshape(-1,1), 4)

    #print('b:',b)
    (bvi,bvx), (bti,btx) = MinSorter(b.reshape(-1,1), 4)

    #print('Results:')
    #print('a:',avx.reshape(-1,),atx.reshape(-1,))
    #print('b:',bvx.reshape(-1,),btx.reshape(-1,))
    for (ai,bi) in zip(avx,bvx):
        assert ai==bi
    for (at,bt) in zip(atx,btx):
        assert at==bt
    for (ai,bi) in zip(avx,b[bvi]):
        assert ai==bi
    for (at,bt) in zip(atx,b[bti]):
        assert at==bt

    print('Passed test 1: 10 numbers')

    a = np.zeros((6,6),dtype=np.int)
    for i in range(6):
        a[i,:] = i
    a = a.reshape((-1,))
    random.shuffle(a)
    (avi,avx), (ati,atx) = MinSorter(a.reshape(-1,1), 6)
    for i in range(6):
        assert i in avx

        #print(np.sum(atx.reshape(-1,)==i))
        assert 5==np.sum(atx.reshape(-1,)==i)
    assert len(avx)==6
    assert len(atx)==30
    #print(avx.reshape(-1,),atx.reshape(-1,))

    print('Passed test 2: 6x6')

def __test_MinSorter2():
    """
    """
    #from itertools import cycle
    a = np.arange(10)
    b = np.arange(10)
    random.shuffle(b)

    print('a:',a)

    (avi,avx), (ati,atx) = MinSorter2(a.reshape(-1,1), 3)

    #print('b:',b)
    (bvi,bvx), (bti,btx) = MinSorter2(b.reshape(-1,1), 3)

    #print('Results:')
    print('a:',avx.reshape(-1,),atx.reshape(-1,))
    #print('b:',bvx.reshape(-1,),btx.reshape(-1,))
    for (ai,bi) in zip(avx,bvx):
        assert ai==bi
    for (at,bt) in zip(atx,btx):
        assert at==bt
    for (ai,bi) in zip(avx,b[bvi]):
        assert ai==bi
    for (at,bt) in zip(atx,b[bti]):
        assert at==bt

    print('Passed test 1: 10 numbers')

    a = np.zeros((6,6),dtype=np.int)
    for i in range(6):
        a[i,:] = i
    a = a.reshape((-1,))
    random.shuffle(a)
    (avi,avx), (ati,atx) = MinSorter2(a.reshape(-1,1), 6)
    for i in range(6):
        assert i in avx

        #print(np.sum(atx.reshape(-1,)==i))
        assert 5==np.sum(atx.reshape(-1,)==i)
    assert len(avx)==6
    assert len(atx)==30
    #print(avx.reshape(-1,),atx.reshape(-1,))
    #print(avi)
    #print(ati)

    print('Passed test 2: 6x6')

    a = np.zeros((6,6),dtype=np.int)
    for i in range(6):
        a[i,:] = i
    a = a.reshape((-1,))
    #random.shuffle(a)
    (avi,avx), (ati,atx) = MinSorter2(a.reshape(-1,1), 6)
    for i in range(6):
        assert i in avx

        #print(np.sum(atx.reshape(-1,)==i))
        assert 5==np.sum(atx.reshape(-1,)==i)
    assert len(avx)==6
    assert len(atx)==30
    #print(avx.reshape(-1,),atx.reshape(-1,))
    print(avi)
    print(ati)

    print('Passed test 3: 6x6')


if __name__ == "__main__":
    #__test_dobscv_SC()
    #__test_FoldSortMin()
    #__test_MinSorter()
    __test_MinSorter2()
    print('\nEnd of Line')
