#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A collection of functions used in creating synthetic datasets

@author: Joakim Linja
"""
import numpy as np
from pathlib import Path
import scipy.stats as scs
from ML_functions import (preprocess,dobscv_SC,MinSorter2)
from scipy.interpolate import interp1d


def complexity00(N=300, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 100)
    Real features: 2
    Fake features: 2
    Targets: 1

    Clean data would be (x in [0,10]):
        f0: 2*x+1
        f1: -0.2*x
        t0: f0+f1
    Slight, individual noise added to f0, f1 and t0.
    Target t0 calculated with clean f0 and f1.
    """
    np.random.seed(seed=600)
    xrng = np.linspace(0,10,num=N)
    #
    f0 = lambda x: 2*x+1
    f1 = lambda x: -0.2*x
    t0 = lambda x: f0(x)+f1(x)

    X = np.zeros((N,4))
    Y = np.zeros((N,1))

    X[:,0] = f0(xrng)
    X[:,1] = f1(xrng)
    Y[:,0] = t0(xrng)

    X[:,0] = X[:,0] + (np.random.rand((N))-0.5)/4
    X[:,1] = X[:,1] + (np.random.rand((N))-0.5)/4
    Y[:,0] = Y[:,0] + (np.random.rand((N))-0.5)/4

    if onlyCorrect:
        X = X[:,0:2]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,2] = np.random.rand((N))
    X[:,3] = np.random.normal(size=(N,))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity01(N=400, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 200)
    Real features: 2
    Fake features: 3
    Targets: 1

    Clean data would be (x in [0,10]):
        f0: 0.25*x
        f1: 0.3*sin(5*x)
        t0: f0+f1
    Slight, individual noise added to f0, f1 and t0.
    Target t0 calculated with f0 and f1.
    """
    np.random.seed(seed=601)
    xrng = np.linspace(0,10,num=N)
    f0 = lambda x: 0.25*x
    f1 = lambda x: 0.3*np.sin(5*x)

    X = np.zeros((N,5))
    Y = np.zeros((N,1))

    X[:,0] = f0(xrng) + (np.random.rand((N))-0.5)/4
    X[:,1] = f1(xrng) + (np.random.rand((N))-0.5)/4
    Y[:,0] = X[:,0] + X[:,1] + (np.random.rand((N))-0.5)/4

    if onlyCorrect:
        X = X[:,0:2]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,2] = np.random.rand((N))
    X[:,3] = np.random.normal(size=(N,))
    X[:,4] = np.sin(xrng) + (np.random.rand((N))-0.5)
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity02(N=500, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 200)
    Real features: 2
    Fake features: 3
    Targets: 1

    Clean data would be (x in [-5,5]):
        f0: norm_distribution(x)
        f1: 2
        t0: f0+f1
    Slight, individual noise added to f0, f1 and t0.
    Target t0 calculated with f0 and f1.
    """
    np.random.seed(seed=602)
    xrng = np.linspace(-5,5,num=N)
    f0 = lambda x: scs.norm.pdf(xrng, 0, 1)
    f1 = lambda _: 2

    X = np.zeros((N,5))
    Y = np.zeros((N,1))

    X[:,0] = f0(xrng) + (np.random.rand((N))-0.5)/4
    X[:,1] = f1(xrng) + (np.random.rand((N))-0.5)/4
    Y[:,0] = X[:,0] + X[:,1] + (np.random.rand((N))-0.5)/4

    if onlyCorrect:
        X = X[:,0:2]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,2] = np.random.rand((N))
    X[:,3] = np.random.normal(size=(N,))
    X[:,4] = np.sin(xrng) + (np.random.rand((N))-0.5)
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity03(N=1000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 200)
    Real features: 2
    Fake features: 3
    Targets: 1

    Clean data would be (x in [-5,5]):
        f0: 1
        f1: 2
        t0: 0.2*f0+0.3*f1
    Slight, individual noise added to f0, f1 and t0.
    Target t0 calculated with clean f0 and f1.
    """
    np.random.seed(seed=603)
    xrng = np.linspace(-5,5,num=N)
    f0 = lambda x: 1
    f1 = lambda _: 2

    X = np.zeros((N,5))
    Y = np.zeros((N,1))

    X[:,0] = f0(xrng) + (np.random.rand((N))-0.5)/4
    X[:,1] = f1(xrng) + (np.random.rand((N))-0.5)/4
    Y[:,0] = f0(xrng) + f1(xrng) + (np.random.rand((N))-0.5)/4

    if onlyCorrect:
        X = X[:,0:2]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,2] = np.random.rand((N))
    X[:,3] = np.random.normal(size=(N,))
    X[:,4] = 2*scs.norm.pdf(xrng, 0, 1) + (np.random.rand((N))-0.5)
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity04(N=2000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 200)
    Real features: 10
    Fake features: 20
    Targets: 1

    Clean data would be (x in [-5,5]):
        f0: 1
        f1: 2
        t0: 0.2*f0+0.3*f1
    Slight, individual noise added to f0, f1 and t0.
    Target t0 calculated with clean f0 and f1.
    """
    np.random.seed(seed=604)
    xrng = np.linspace(-5,5,num=N)
    f0 = lambda x: 3*scs.norm.pdf(xrng, -3, 1)+scs.norm.pdf(xrng, 3, 1)
    f1 = lambda x: 2*scs.norm.pdf(xrng, 0, 1)
    f2 = lambda x: np.sin(15*x)
    f3 = lambda x: 2*np.cos(0.2*x+0.2)
    f4 = lambda _: -3
    f5 = lambda _: -2
    f6 = lambda _: -1
    f7 = lambda x: np.sqrt(5+np.abs(x))/np.pi
    f8 = lambda x: 0.7*np.cos(20*np.abs(x))
    f9 = lambda x: (1/20)*np.power(x,2) -0.2*x + np.pi/3

    X = np.zeros((N,30))
    Y = np.zeros((N,1))

    vari = 1e-3
    X[:,0] = f0(xrng) + np.random.normal(scale=vari,size=(N,))
    X[:,1] = f1(xrng) + np.random.normal(scale=vari,size=(N,))
    X[:,2] = f2(xrng) + np.random.normal(scale=vari,size=(N,))
    X[:,3] = f3(xrng) + np.random.normal(scale=vari,size=(N,))
    X[:,4] = f4(xrng) + np.random.normal(scale=0.5, size=(N,))
    X[:,5] = f5(xrng) + np.random.normal(scale=0.75,size=(N,))
    X[:,6] = f6(xrng) + np.random.normal(scale=1, size=(N,))
    X[:,7] = f7(xrng) + np.random.normal(scale=vari,size=(N,))
    X[:,8] = f8(xrng) + np.random.normal(scale=vari,size=(N,))
    X[:,9] = f9(xrng) + np.random.normal(scale=vari,size=(N,))

    Y[:,0] = np.sum(X[:,0:10], axis=1)

    if onlyCorrect:
        X = X[:,0:10]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    # indices 10-14 are based on random.rand
    X[:,10:15] = np.random.rand(N,5)
    # indices 15-19 are based on normal distribution
    X[:,15:20] = np.random.normal(size=(N,5))
    # indices 20-24 are based on poisson distribution
    X[:,20:25] = np.random.poisson(size=(N,5))
    # indices 25-29 are trigonometric with noise
    X[:,25] = 0.2*np.sin(xrng) + np.random.normal(scale=1e-2,size=(N,))
    X[:,26] = 0.1*np.cos(100*xrng+np.random.normal(scale=5e-1,size=(N,)))
    X[:,27] = np.sin(np.random.rand(N))
    X[:,28] = np.cos(np.random.normal(size=(N,)))
    X[:,29] = np.sin(xrng)+0.5*np.cos(2*xrng)+np.random.normal(size=(N,))



    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity05(N=1500, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 1500)
    Real features: 5
    Fake features: 5
    Targets: 1
    """
    np.random.seed(seed=605)
    xrng = np.linspace(1e-2,10,num=N)

    X = np.zeros((N,10))
    Y = np.zeros((N,1))

    X[:,0] = 1/np.sqrt(xrng) + np.random.normal(scale=0.2,size=(N,))
    X[:,1] = 1+0.5*np.cos(0.1*xrng) + np.random.normal(scale=0.1,size=(N,))
    X[:,2] = 5*scs.norm.pdf(xrng, 0, 2) + np.random.normal(scale=0.2,size=(N,))
    X[:,3] = 5*scs.norm.pdf(xrng, 5, 2) + np.random.normal(scale=0.2,size=(N,))
    X[:,4] = 5*scs.norm.pdf(xrng, 10, 2) + np.random.normal(scale=0.2,size=(N,))

    Y[:,0] = np.sum(X[:,0:5],axis=1)

    if onlyCorrect:
        X = X[:,0:5]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,5:] = np.random.normal(size=(N,5))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity06(N=2000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 2000)
    Real features: 6
    Fake features: 6
    Targets: 1
    """
    np.random.seed(seed=606)
    xrng = np.linspace(0,100,num=N)

    X = np.zeros((N,12))
    Y = np.zeros((N,1))

    X[:,0] = np.cos(xrng) + np.random.normal(scale=0.1,size=(N,))
    X[:,1] = np.cos(2*xrng) + np.random.normal(scale=0.1,size=(N,))
    X[:,2] = np.cos(4*xrng) + np.random.normal(scale=0.1,size=(N,))
    X[:,3] = np.cos(8*xrng) + np.random.normal(scale=0.1,size=(N,))
    X[:,4] = np.cos(16*xrng) + np.random.normal(scale=0.1,size=(N,))
    X[:,5] = np.cos(32*xrng) + np.random.normal(scale=0.1,size=(N,))

    Y[:,0] = X[:,0]-X[:,1]+X[:,2]-X[:,3]+X[:,4]-X[:,5]

    if onlyCorrect:
        X = X[:,0:6]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,6] = np.cos(xrng-0.9*np.pi) + np.random.normal(scale=0.1,size=(N,))
    X[:,7] = np.cos(2*xrng+0.1*np.pi) + np.random.normal(scale=0.1,size=(N,))
    X[:,8] = np.cos(4*xrng-0.1*np.pi) + np.random.normal(scale=0.1,size=(N,))
    X[:,9] = np.cos(8*xrng+0.2*np.pi) + np.random.normal(scale=0.1,size=(N,))
    X[:,10] = np.cos(16*xrng-0.2*np.pi) + np.random.normal(scale=0.1,size=(N,))
    X[:,11] = np.cos(32*xrng+0.5*np.pi) + np.random.normal(scale=0.1,size=(N,))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def _trianglewave(x):

    xi = len(x)

    N = 5
    I = 1
    n = 2*I+1
    f = 440
    res = np.zeros((xi,N))
    for i in range(N):
        res[:,i] = np.power(-1,i)*np.power(float(n),-2)*np.sin(2*np.pi*f*n*x)
    res = np.sum(res,axis=1)

    res = (8/(np.pi**2))*res
    return res



def complexity07(N=2000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 200)
    Real features: 6
    Fake features: 0
    Targets: 1
    All values are necessary.
    """
    np.random.seed(seed=607)
    xrng = np.linspace(-5,5,num=N)

    X = np.zeros((N,6))
    Y = np.zeros((N,1))

    X[:,0] = -0.5*xrng
    X[:,1] = 0.1*np.power(xrng,2)
    X[:,2] = np.sin(xrng)
    X[:,3] = 3*scs.norm.pdf(xrng, loc=0, scale=1)
    X[:,4] = 3*_trianglewave(xrng)
    X[:,5] = -(1/120)*np.power(xrng,3)

    Y[:,0] = np.sum(X,axis=1)

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity08(N=2000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 2000)
    Real features: 6
    Fake features: 6
    Targets: 1
    """
    np.random.seed(seed=608)
    x = np.linspace(0,10,num=N)

    X = np.zeros((N,12))
    Y = np.zeros((N,1))

    def f1(z):
        if z<=2:
            return -1
        elif 2 < z and z<= 4:
            return 0
        elif 4 < z and z <= 6:
            return 1
        elif 6 < z and z <= 8:
            return 1/3
        else:
            return -1/3

    X[:,0] = np.array(list(map(lambda z: f1(z), x)))
    X[:,1] = np.heaviside(x,-0.5)
    X[:,2] = np.round(10*np.sin(x)).astype(np.int)/10
    X[:,3] = 3*scs.norm.pdf(x,loc=5, scale=1)
    X[:,4] = np.sqrt(np.abs(x))
    X[:,5] = 0.2*np.sin(440*x)

    X[:,0:6] = X[:,0:6] + np.random.normal(scale=0.15,size=(N,6))

    Y[:,0] = np.sum(X,axis=1)

    if onlyCorrect:
        X = X[:,0:6]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,6:9] = np.random.randint(0,high=10,size=(N,3))
    X[:,9:12] = np.random.triangular(3,6,8,size=(N,3))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity09(N=2000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 2000)
    Real features: 8
    Fake features: 16
    Targets: 1
    """
    np.random.seed(seed=609)
    x = np.linspace(0,10,num=N)

    X = np.zeros((N,24))
    Y = np.zeros((N,1))

    def f1(z):
        if z<=2:
            return -1
        elif 2 < z and z<= 4:
            return 0
        elif 4 < z and z <= 6:
            return 1
        elif 6 < z and z <= 8:
            return 1/3
        else:
            return -1/3

    def f2(z):
        if z<=1:
            return 5
        elif 1< z and z <=2 :
            return 3
        elif 2< z and z <=3 :
            return 1
        elif 3< z and z <=4 :
            return -1
        elif 4< z and z <=5 :
            return 4
        elif 5< z and z <=6 :
            return 2
        elif 6< z and z <=7 :
            return 0
        elif 7< z and z <=8 :
            return 3
        elif 8< z and z <=9 :
            return 1
        else:
            return -1

    def f3(z):
        if z <= 5:
            return 0
        else:
            return -1

    X[:,0] = np.array(list(map(lambda z: f1(z), x)))
    X[:,1] = np.array(list(map(lambda z: f2(z), x)))
    X[:,2] = np.array(list(map(lambda z: f3(z), x)))
    X[:,3] = np.heaviside(x,-0.5)
    X[:,4] = np.round(10*np.sin(0.5*x)).astype(np.int)/10
    X[:,5] = 2*scs.norm.pdf(x,loc=10, scale=1)
    X[:,6] = 0.2*np.cos(0.5*x)
    X[:,7] = 0.2*np.sin(440*x)

    X[:,0:8] = X[:,0:8] + np.random.normal(scale=0.15,size=(N,8))

    Y[:,0] = np.sum(X,axis=1)
    Y[:,0] = (100*Y[:,0]).astype(np.int)

    if onlyCorrect:
        X = X[:,0:8]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,8:15] = np.random.randint(0,high=10,size=(N,7))
    X[:,15:17] = np.random.triangular(0,1,5,size=(N,2))
    X[:,17:19] = np.random.triangular(4,5,6,size=(N,2))
    X[:,19:21] = np.random.triangular(5,9,10,size=(N,2))
    X[:,21:24] = np.random.normal(scale=1,size=(N,3))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def F(n):
        """
        Piano string frequency
        """
        return 440*np.power(np.power(2,1/12), n-49)


def complexity10(N=4000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 4000)
    Real features: 15
    Fake features: 385
    Targets: 1
    """
    np.random.seed(seed=610)
    x = np.linspace(-5,5,num=N)
    xf = 2*np.pi*np.linspace(-5,5,num=N)
    X = np.zeros((N,400))
    Y = np.zeros((N,1))

    X[:,0] = -0.5*x
    X[:,1] = 0.1*np.power(x,2)
    X[:,2] = np.sin(xf)
    X[:,3] = 3*scs.norm.pdf(x, loc=0, scale=1)
    X[:,4] = 3*_trianglewave(x)
    X[:,5] = -(1/120)*np.power(x,3)
    X[:,6]  = np.sin(xf*F(1))
    X[:,7]  = np.sin(xf*F(3))
    X[:,8]  = np.sin(xf*F(5))
    X[:,9]  = np.sin(xf*F(7))
    X[:,10] = np.sin(xf*F(9))
    X[:,11] = np.sin(xf*F(11))
    X[:,12] = np.sin(xf*F(13))
    X[:,13] = np.sin(xf*F(15))
    X[:,14] = np.sin(xf*F(17))

    Y[:,0] = np.sum(X[:,0:6],axis=1)
    Y[:,0] = Y[:,0]+0.5*X[:,6]+0.5*X[:,7]+0.5*X[:,8]+0.5*X[:,9]+0.5*X[:,10]
    Y[:,0] = Y[:,0]+0.5*X[:,11]+0.5*X[:,12]+0.5*X[:,13]+0.5*X[:,14]

    if onlyCorrect:
        X = X[:,0:15]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)


    N11 = (N,11)
    X[:,15:26]   = np.random.normal(size=N11)
    X[:,26:37]   = np.random.chisquare(1,size=N11)
    X[:,37:48]   = np.random.beta(1,2,size=N11)
    X[:,48:59]   = np.random.binomial(10,0.5,size=N11)
    X[:,59:70]   = np.random.dirichlet([11],size=N11).reshape(-1,11)
    X[:,70:81]   = np.random.exponential(2,size=N11)
    X[:,81:92]   = np.random.f(3,3,size=N11)
    X[:,92:94]   = np.random.gamma((11,3),size=(N,2))
    X[:,94]      = np.sin(xf*F(2))
    X[:,95]      = np.sin(xf*F(4))
    X[:,96]      = np.sin(xf*F(6))
    X[:,97]      = np.sin(xf*F(8))
    X[:,98]      = np.sin(xf*F(10))
    X[:,99]      = np.sin(xf*F(12))
    X[:,100]     = np.sin(xf*F(14))
    X[:,101]     = np.sin(xf*F(16))
    X[:,102]     = np.sin(xf*F(18))
    X[:,94:103]  = X[:,94:103] + np.random.normal(scale=0.1,size=(N,9))
    X[:,103:114] = np.random.geometric(0.5,size=N11)
    X[:,114:125] = np.random.gumbel(size=N11)
    X[:,125:136] = np.random.hypergeometric(10,5,10,size=N11)
    X[:,136:147] = np.random.laplace(size=N11)
    X[:,147:158] = np.random.logistic(size=N11)
    X[:,158:169] = np.random.lognormal(size=N11)
    X[:,169:180] = np.random.logseries(0.5,size=N11)
    X[:,180:191] = np.random.multinomial(10,[0.5],size=N11).reshape(-1,11)
    X[:,191:202] = np.random.multivariate_normal([2],[[1]],size=N11).reshape(-1,11)
    X[:,202:213] = np.random.negative_binomial(10,0.5,size=N11)
    X[:,213:224] = np.random.noncentral_chisquare(1,2,size=N11)
    X[:,224:235] = np.random.noncentral_f(3,3,25,size=N11)
    X[:,235:237] = np.random.pareto((7,13),size=(N,2))
    X[:,237:246] = np.random.normal(size=(N,9))
    X[:,246:257] = np.random.poisson(size=N11)
    X[:,257:268] = np.random.power(2,size=N11)
    X[:,268:279] = np.random.rayleigh(size=N11)
    X[:,279:290] = np.random.standard_cauchy(size=N11)
    X[:,290:301] = np.random.standard_exponential(size=N11)
    X[:,301:303] = np.random.standard_gamma((11,3),size=(N,2))
    X[:,303:312] = np.random.normal(size=(N,9))
    X[:,312:323] = np.random.standard_normal(size=N11)
    X[:,323:334] = np.random.standard_t(4,size=N11)
    X[:,334:345] = np.random.triangular(-5,-1,3,size=N11)
    X[:,345:356] = np.random.uniform(size=N11)
    X[:,356:367] = np.random.vonmises(0,1e-2,size=N11)
    X[:,367:378] = np.random.wald(1.5,3,size=N11)
    X[:,378:380] = np.random.weibull((13,7),size=(N,2))
    X[:,380:389] = np.random.normal(size=(N,9))
    X[:,389:400] = np.random.zipf(5,size=N11)

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity11(N=3000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 3000)
    Real features: 10
    Fake features: 10
    Targets: 1
    """
    np.random.seed(seed=611)
    x = np.linspace(0,500,num=N)
    xf = 2*np.pi*x
    X = np.zeros((N,20))
    Y = np.zeros((N,1))


    X[:,0]  = np.sin(xf*F(1))#
    X[:,1]  = np.sin(xf*F(3))#
    X[:,2]  = np.sin(xf*F(5))#
    X[:,3]  = np.sin(xf*F(7))#
    X[:,4]  = np.sin(xf*F(9))#
    X[:,5]  = np.sin(xf*F(11))#
    X[:,6]  = np.sin(xf*F(13))#
    X[:,7]  = np.sin(xf*F(15))#
    X[:,8]  = np.sin(xf*F(17))#
    X[:,9]  = np.sin(xf*F(19))#

    Y[:,0] = Y[:,0] + X[:,0]*X[:,5] - X[:,3]*X[:,7]
    Y[:,0] = Y[:,0] + np.sin(xf*(X[:,1]+X[:,4]+X[:,6]+X[:,8]))
    Y[:,0] = Y[:,0] - np.sin(xf*X[:,2])
    Y[:,0] = Y[:,0] + np.sin(xf*F(((X[:,9]+2)*50).astype(np.int)))

    if onlyCorrect:
        X = X[:,0:10]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,10] = np.sin(xf*F(2))
    X[:,11] = np.sin(xf*F(4))
    X[:,12] = np.sin(xf*F(6))
    X[:,13] = np.sin(xf*F(8))
    X[:,14] = np.sin(xf*F(10))
    X[:,15] = np.sin(xf*F(12))
    X[:,16] = np.sin(xf*F(14))
    X[:,17] = np.sin(xf*F(16))
    X[:,18] = np.sin(xf*F(18))
    X[:,19] = np.sin(xf*F(20))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity12(N=3000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 3000)
    Real features: 10
    Fake features: 10
    Targets: 1
    """
    np.random.seed(seed=612)
    x = np.linspace(0,500,num=N)
    xf = 2*np.pi*x
    X = np.zeros((N,20))
    Y = np.zeros((N,1))


    X[:,0]  = np.sin(xf*F(1))#
    X[:,1]  = np.sin(xf*F(3))#
    X[:,2]  = np.sin(xf*F(5))#
    X[:,3]  = np.sin(xf*F(7))#
    X[:,4]  = np.sin(xf*F(9))#
    X[:,5]  = np.sin(xf*F(11))#
    X[:,6]  = np.sin(xf*F(13))#
    X[:,7]  = np.sin(xf*F(15))#
    X[:,8]  = np.sin(xf*F(17))#
    X[:,9]  = np.sin(xf*F(19))#

    X[:,0:10] = X[:,0:10] + np.random.normal(scale=1e-1, size=(N,10))

    Y[:,0] = Y[:,0] + X[:,0]*X[:,5] - X[:,3]*X[:,7]
    Y[:,0] = Y[:,0] + np.sin(xf*(X[:,1]+X[:,4]+X[:,6]+X[:,8]))
    Y[:,0] = Y[:,0] - np.sin(xf*X[:,2])
    Y[:,0] = Y[:,0] + np.sin(xf*F(((X[:,9]+2)*50).astype(np.int)))

    if onlyCorrect:
        X = X[:,0:10]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,10] = np.sin(xf*F(2))
    X[:,11] = np.sin(xf*F(4))
    X[:,12] = np.sin(xf*F(6))
    X[:,13] = np.sin(xf*F(8))
    X[:,14] = np.sin(xf*F(10))
    X[:,15] = np.sin(xf*F(12))
    X[:,16] = np.sin(xf*F(14))
    X[:,17] = np.sin(xf*F(16))
    X[:,18] = np.sin(xf*F(18))
    X[:,19] = np.sin(xf*F(20))

    X[:,10:20] = X[:,10:20] + np.random.normal(scale=1e-1, size=(N,10))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity13(N=3000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 3000)
    Real features: 10
    Fake features: 10
    Targets: 1
    """
    np.random.seed(seed=613)
    x = np.linspace(0,500,num=N)
    xf = 2*np.pi*x
    X = np.zeros((N,20))
    Y = np.zeros((N,1))

    X[:,0]  = np.sin(xf*F(1))
    X[:,1]  = np.sin(xf*F(3))
    X[:,2]  = np.sin(xf*F(5))
    X[:,3]  = np.sin(xf*F(7))
    X[:,4]  = np.sin(xf*F(9))
    X[:,5]  = np.sin(xf*F(11))
    X[:,6]  = np.sin(xf*F(13))
    X[:,7]  = np.sin(xf*F(15))
    X[:,8]  = np.sin(xf*F(17))
    X[:,9]  = np.sin(xf*F(19))

    Y[:,0] = np.sum(X[:,0:10],axis=1)

    if onlyCorrect:
        X = X[:,0:10]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,10] = np.sin(xf*F(2))
    X[:,11] = np.sin(xf*F(4))
    X[:,12] = np.sin(xf*F(6))
    X[:,13] = np.sin(xf*F(8))
    X[:,14] = np.sin(xf*F(10))
    X[:,15] = np.sin(xf*F(12))
    X[:,16] = np.sin(xf*F(14))
    X[:,17] = np.sin(xf*F(16))
    X[:,18] = np.sin(xf*F(18))
    X[:,19] = np.sin(xf*F(20))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def complexity14(N=3000, unModOrig=False, onlyCorrect=False):
    """
    Observations: N (default 3000)
    Real features: 10
    Fake features: 10
    Targets: 1
    """
    np.random.seed(seed=614)
    x = np.linspace(0,500,num=N)
    xf = 2*np.pi*x
    X = np.zeros((N,20))
    Y = np.zeros((N,1))

    X[:,0]  = np.sin(xf*F(1))
    X[:,1]  = np.sin(xf*F(3))
    X[:,2]  = np.sin(xf*F(5))
    X[:,3]  = np.sin(xf*F(7))
    X[:,4]  = np.sin(xf*F(9))
    X[:,5]  = np.sin(xf*F(11))
    X[:,6]  = np.sin(xf*F(13))
    X[:,7]  = np.sin(xf*F(15))
    X[:,8]  = np.sin(xf*F(17))
    X[:,9]  = np.sin(xf*F(19))

    X[:,0:10] = X[:,0:10] + np.random.normal(scale=0.15,size=(N,10))
    Y[:,0] = np.sum(X[:,0:10],axis=1)

    if onlyCorrect:
        X = X[:,0:10]
        return preprocess(X,Y, unmodifiedOriginals=unModOrig)

    X[:,10] = np.sin(xf*F(2))
    X[:,11] = np.sin(xf*F(4))
    X[:,12] = np.sin(xf*F(6))
    X[:,13] = np.sin(xf*F(8))
    X[:,14] = np.sin(xf*F(10))
    X[:,15] = np.sin(xf*F(12))
    X[:,16] = np.sin(xf*F(14))
    X[:,17] = np.sin(xf*F(16))
    X[:,18] = np.sin(xf*F(18))
    X[:,19] = np.sin(xf*F(20))

    X[:,10:20] = X[:,10:20] + np.random.normal(scale=0.15,size=(N,10))

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def ArtDat000(N=1000, unModOrig=False):
    """
    """
    np.random.seed(seed=1000)
    x = np.linspace(0,5,num=N)
    # Features
    X = np.zeros((N,2))
    Y = np.zeros((N,1))
    # Real features
    X[:,0] = x*x+1
    # Fake features
    X[:,1] = np.pi*x
    # Output
    Y = 2*X[:,0]
    Y = Y.reshape(-1,1)
    # Simulated measurement errors
    rnd = np.random.normal(scale=0.15,size=(N,3))
    X[:,:] += rnd[:,0:2]
    Y[:,0] += rnd[:,2]

    return X,Y
    #return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def ArtDat001(N=1000, unModOrig=False):
    """
    """

    def f(a):
        return np.array(list(map(lambda b: (8/2)*b+5 if b <= 0 else (-8/2)*b+5,a)))

    np.random.seed(seed=1001)
    x = np.linspace(-2,2,num=N)
    # Features
    X = np.zeros((N,2))
    Y = np.zeros((N,1))
    # Real features
    X[:,0] = -2*x*x+5
    # Fake features
    X[:,1] = f(x)
    # Output
    Y = (X[:,0]*-1).reshape(-1,1)

    # Simulated measurement errors
    rnd = np.random.normal(scale=0.15,size=(N,3))
    X[:,:] += rnd[:,0:2]
    Y[:,0] += rnd[:,2]

    #return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def ArtDat002(N=1000, unModOrig=False):
    """
    """
    np.random.seed(seed=1002)
    x = np.linspace(-np.pi,np.pi,num=N)
    # Featues
    X = np.zeros((N,3))
    Y = np.zeros((N,1))
    # Real features
    X[:,0] = np.sin(x)
    X[:,1] = np.sin(x*np.random.normal(loc=1,scale=0.1,size=(N)))
    # Fake features
    xf = np.linspace(-np.pi,np.pi,num=7)
    ff = interp1d(xf, np.sin(xf))
    X[:,2] = ff(x)

    Y[:,0] = np.sum(X[:,:],axis=1)

    #return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def ArtDat003(N=1000, unModOrig=False): # Nightmarish
    """
    Climate warming based.

    Breaks filter algorithm.
    """
    np.random.seed(seed=1003)
    x = np.linspace(2*np.pi,7.25*np.pi,num=N)
    # Features
    X = np.zeros((N,4))
    Y = np.zeros((N,1))
    # Real features
    nrmlr = np.random.normal(loc=0,scale=0.005,size=(N))
    #print(nrmlr)
    freqs = np.ones((N))
    f = 1
    for i in range(N):
        f += nrmlr[i]
        freqs[i] = f

    for i in range(N):
        X[i,0] = np.sin(freqs[i]*x[i])

    X[:,1] = 1.5*np.sqrt(x)-0.025*x
    #X[:,1] = 0.5*np.sqrt(x)-0.025*x
    # Fake features
    X[:,2] = np.sin(2.55*np.pi+2*x)
    X[:,3] = (-1.5*np.sqrt(5*x)-2.575*x**8+1.2)/140000000000 #*x**6+1.2)/140000000
    #X[:,3] = 0.5*np.sqrt(0.9*x)-0.075*x+1.2

    # Output
    for i in range(N):
        Y[i,0] = np.cos(X[i,0])*X[i,1]

    # Simulation of measurement errors
    #X[:,:] += np.random.normal(loc=0,scale=0.05,size=(N,4))
    #Y[:,:] += np.random.normal(loc=0,scale=0.05,size=(N,1))

    #return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def ArtDat004(N=1000, unModOrig=False):
    """
    """
    np.random.seed(seed=1004)
    x = np.linspace(2*np.pi,8*np.pi,num=N)
    x2 = np.linspace(2*np.pi,8*np.pi,num=2*N)

    nrnd = np.random.normal(loc=0,scale=0.01,size=(2*N,))
    inrnd = interp1d(x2, nrnd)
    #print(np.shape(inrnd))

    # Featues
    X = np.zeros((N,2))
    Y = np.zeros((N,1))
    # Real features
    #X[:,0] = nrnd#np.sin(x)
    a = inrnd(x2[0::2])

    X[:,0] = inrnd(x2[0::2]) #np.sin(x*np.random.normal(loc=0,scale=0.1,size=(N)))
    X[:,1] = inrnd(x2[1::2])
    # Fake features
    #xf = np.linspace(-np.pi,np.pi,num=7)
    #ff = interp1d(xf, np.sin(xf))
    #X[:,2] = np.random.normal(loc=0,scale=0.05,size=(N,1))

    Y[:,0] = np.random.normal(loc=0,scale=0.0005,size=(N,))#np.sum(X[:,:],axis=1)

    #return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)



def ArtDat005(N=1000, unModOrig=False):
    """
    """
    np.random.seed(seed=1005)

    x = np.linspace(0,10*np.pi,num=N)
    xd = np.linspace(0,N,num=N)

    # Featues
    X = np.zeros((N,3))
    Y = np.zeros((N,1))

    a = -1/(np.pi+10*np.pi**2)
    b = -(3+38*np.pi)/(np.pi+10*np.pi**2)
    c = 10


    X[:,0] = np.sin(x*(a*x**2+b*x+c))
    X[:,1] = x*(a*x**2+b*x+c)
    X[:,2] = (a*x**2+b*x+c)


    Y[:,0] = np.random.normal(loc=0,scale=0.0005,size=(N,))#np.sum(X[:,:],axis=1)

    #return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth100_lin(N=1000, fN=100, noise=False, unModOrig=False):
    np.random.seed(seed=261010)
    X = np.zeros((N,fN))
    Y = np.zeros((N,1))

    X = np.random.random((N,fN))

    for i in range(100):
        Y[:,0] += (99-i)*X[:,i]

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    #return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth100_sin(N=1000, fN=100, noise=False, unModOrig=False):
    """
    """
    np.random.seed(seed=261011)
    X = np.zeros((N,fN))
    Y = np.zeros((N,1))
    X = np.random.random((N,fN))

    for i in range(100):
        Y[:,0] += np.sin(2*np.pi*(99-i)*X[:,i])

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth100_pow2(N=1000, fN=100, noise=False, unModOrig=False):
    """
    """
    np.random.seed(seed=261012)
    X = np.zeros((N,fN))
    Y = np.zeros((N,1))
    X = np.random.random((N,fN))

    for i in range(100):
        Y[:,0] += (99-i)*np.power(X[:,i],2)

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth100_pow6(N=1000, fN=100, noise=False, unModOrig=False):
    """
    """
    np.random.seed(seed=261013)
    X = np.zeros((N,fN))
    Y = np.zeros((N,1))
    X = np.random.random((N,fN))

    for i in range(100):
        Y[:,0] += (99-i)*np.power(X[:,i],6)

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth100_exp(N=1000, fN=100, noise=False, unModOrig=False):
    """
    """
    np.random.seed(seed=261014)
    X = np.zeros((N,fN))
    Y = np.zeros((N,1))
    X = np.random.random((N,fN))

    for i in range(100):
        Y[:,0] += (99-i)*np.exp(X[:,i])

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth100_log(N=1000, fN=100, noise=False, unModOrig=False):
    """
    """
    np.random.seed(seed=261015)
    X = np.zeros((N,fN))
    Y = np.zeros((N,1))
    X = np.random.random((N,fN))

    for i in range(100):
        Y[:,0] += (99-i)*np.log(1-X[:,i])
        # 1-X because random is [0,1) so now log(0) is avoided.

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth_ref_1(N=1000, fN=100, noise=False, unModOrig=False):
    """
    Feature Selection for Nonlinear Regression and its Application to Cancer Research
    Eq 4.18
    """
    np.random.seed(seed=261016)
    X = np.random.random((N,fN))
    Y = np.zeros((N,1))

    Y[:,0] += -2*np.sin(2*X[:,0])+np.power(X[:,1],2)+X[:,2]+np.exp(X[:,3])

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth_ref_2(N=1000, fN=100, noise=False, unModOrig=False):
    """
    Feature Selection for Nonlinear Regression and its Application to Cancer Research
    Eq 4.19
    """
    np.random.seed(seed=261017)
    X = np.random.random((N,fN))
    Y = np.zeros((N,1))

    Y[:,0] += X[:,0]*np.exp(2*X[:,1]) + np.power(X[:,2],2)

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth_ref_3(N=1000, fN=100, noise=False, unModOrig=False):
    """
    Feature Selection for Nonlinear Regression and its Application to Cancer Research
    Eq 4.20
    """
    np.random.seed(seed=261018)
    X = np.random.random((N,fN))*4
    Y = np.zeros((N,1))


    Y[:,0] += np.sin(2*np.pi*X[:,0])

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=0.1,size=(N,))
        #n = n/np.ptp(n)
        Y[:,0] += n

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def synth_ref_4(N=1000, fN=100, noise=False, unModOrig=False):
    """
    Feature Selection for Nonlinear Regression and its Application to Cancer Research
    Eq 4.21
    """
    np.random.seed(seed=261019)
    X = np.random.random((N,fN))
    Y = np.random.random((N,1))*20

    X[:,0] = Y[:,0]*np.sin(Y[:,0])
    X[:,1] = Y[:,0]*np.cos(Y[:,0])

    if noise:
        # Zero mean and unit variance
        n = np.random.normal(loc=0,scale=1,size=(N,2))
        #n = n/np.ptp(n)
        X[:,0] += n[:,0]
        X[:,1] += n[:,1]

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def correctSynDataFeatures(featN) -> dict:
    """
    """
    AdatZN = featN-100
    if AdatZN <0:
        AdatZN = 0
    Adat = np.hstack((np.ones(100,dtype=np.int),np.zeros(AdatZN,dtype=np.int)))
    res = {
        'R1':np.hstack((np.ones(4,dtype=np.int),np.zeros(featN-4,dtype=np.int))), # 4
        'R2':np.hstack((np.ones(3,dtype=np.int),np.zeros(featN-3,dtype=np.int))), # 3
        'R3':np.hstack((np.ones(1,dtype=np.int),np.zeros(featN-1,dtype=np.int))), # 1
        'R4':np.hstack((np.ones(2,dtype=np.int),np.zeros(featN-2,dtype=np.int))), # 2
        'A1':Adat,
        'A2':Adat,
        'A3':Adat,
        'A4':Adat,
        'A5':Adat,
        'A6':Adat,
        }
    return res


def savePreprocessedDatasets(dskeys):
    """
    """
    datasets = getArtificialDatasets()

    dsf = Path('Experiments/Saved_datasets')
    parf = Path(dsf,'Params')

    def fpth(pth,add):
        return Path(pth,add).as_posix()

    Nls = [200,500,1000,1500,2000,2500,3000]
    for N in Nls:
        for dskey in dskeys:
            df = datasets[dskey]
            res = df(N=N,unModOrig=True)
            X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask = res


            #if dskey=='SCM1D':
            #    Y = Y[:,0]
            #    Y = Y.reshape((-1,1))
            #    nY = Yscaler.Scaling(Y)

            np.save(fpth(dsf,'{0:s}_{1:d}_origX.npy').format(dskey,N),X)
            np.save(fpth(dsf,'{0:s}_{1:d}_origY.npy').format(dskey,N),Y)
            np.save(fpth(dsf,'{0:s}_{1:d}_normX.npy').format(dskey,N),nX)
            np.save(fpth(dsf,'{0:s}_{1:d}_normY.npy').format(dskey,N),nY)
            (xmin,xmax) = Xscaler.getScalingParams()
            (ymin,ymax) = Yscaler.getScalingParams()
            np.save(fpth(parf,'{0:s}_{1:d}_xmin.npy').format(dskey,N),xmin)
            np.save(fpth(parf,'{0:s}_{1:d}_xmax.npy').format(dskey,N),xmax)
            np.save(fpth(parf,'{0:s}_{1:d}_ymin.npy').format(dskey,N),ymin)
            np.save(fpth(parf,'{0:s}_{1:d}_ymax.npy').format(dskey,N),ymax)
            np.save(fpth(parf,'{0:s}_{1:d}_xconstmask.npy').format(dskey,N),
                    XconstFiltMask)
            np.save(fpth(parf,'{0:s}_{1:d}_yconstmask.npy').format(dskey,N),
                    YconstFiltMask)

            print(dskey)
            print(np.shape(nX))
            print(np.shape(nY))


def CreateSyntheticDatasets(N=1000, noise=True):
    """
    """
    synthsets = {
        'synth_R1':synth_ref_1,
        'synth_R2':synth_ref_2,
        'synth_R3':synth_ref_3,
        'synth_R4':synth_ref_4,
        'synth_A1':synth100_lin,
        'synth_A2':synth100_sin,
        'synth_A3':synth100_pow2,
        'synth_A4':synth100_pow6,
        'synth_A5':synth100_exp,
        'synth_A6':synth100_log,
        }

    def fpth(pth,add):
        return Path(pth,add).as_posix()

    dsf = Path('Data/Saved_datasets3')
    parf = Path(dsf,'Params')

    #N = 1000
    #noise = True

    # (N=1000, fN=100, noise=False, unModOrig=False)
    # Synthset:
    for dkey in synthsets:
        print('')
        print('Dataset:',dkey)
        fun = synthsets[dkey]
        # number of features:
        for p in [200]:# [100,200,2000]:
            print('Handling feature number:',p)
            # Form dataset
            X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask = \
                fun(N=N,fN=p,noise=True,unModOrig=False)

            np.save(fpth(dsf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_origX.npy').format(dkey,N,p,noise),X)
            np.save(fpth(dsf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_origY.npy').format(dkey,N,p,noise),Y)
            np.save(fpth(dsf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_normX.npy').format(dkey,N,p,noise),nX)
            np.save(fpth(dsf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_normY.npy').format(dkey,N,p,noise),nY)
            (xmin,xmax) = Xscaler.getScalingParams()
            (ymin,ymax) = Yscaler.getScalingParams()
            np.save(fpth(parf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_xmin.npy').format(dkey,N,p,noise),xmin)
            np.save(fpth(parf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_xmax.npy').format(dkey,N,p,noise),xmax)
            np.save(fpth(parf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_ymin.npy').format(dkey,N,p,noise),ymin)
            np.save(fpth(parf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_ymax.npy').format(dkey,N,p,noise),ymax)
            np.save(fpth(parf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_xconstmask.npy').format(dkey,N,p,noise),
                    XconstFiltMask)
            np.save(fpth(parf,'{0:s}_N{1:d}_p{2:d}_ns{3:d}_yconstmask.npy').format(dkey,N,p,noise),
                    YconstFiltMask)


            TrainValPreSplit(dkey, nX, nY, Xscaler, Yscaler, XconstFiltMask, outputSel=-1,
                 separateValidationSet=False,
                 validationSetP=0.33,validationSet=None,
                 deterministicDobSCV=False,
                 forceTrainValSplit=True,
                 titlevar='_N{0:d}_p{1:d}_ns{2:d}'.format(N,p,noise))



def TrainValPreSplit(dkey, X, Y, xscaler, yscaler, XconstFiltMask, outputSel=-1,
                 separateValidationSet=False,
                 validationSetP=0.33,validationSet=None,
                 deterministicDobSCV=False,
                 forceTrainValSplit=True,
                 titlevar=''):
    """
    Includes the ability to force own train/val split
    """
    print('TrainValPreSplit')

    #print(np.shape(XconstFiltMask))
    #print(np.sum(XconstFiltMask))

    #raise NotImplementedError('Intentional breakpoint')

    oX = xscaler.convertFromScale(X)
    oY = yscaler.convertFromScale(Y)

    (N,xm) = np.shape(oX)
    valXconstFiltMask = None
    valYconstFiltMask = None

    if outputSel != -1:
        oY = oY[:,outputSel]
        oY = oY.reshape((N,1))

    for itrn in range(0,1):#40):

        if type(validationSet) is tuple and not forceTrainValSplit:
            raise NotImplementedError

        #if validationSet!=tuple:
        else:
            print('Taking validation dataset from training data with FoldSorter')
            if type(validationSet) is tuple and forceTrainValSplit:
                print('There is a readily provided validation dataset.')
                print('Combining it with the training data.')
                (valX, valY) = validationSet
                valN = len(valX)
                #print(valY)

                """
                validation set converter:
                (X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask) = trainOut
                vX = valX[:,XconstFiltMask]
                vX = Xscaler.convertToScale(vX)
                vY = valY[:,YconstFiltMask]
                vY = Yscaler.convertToScale(vY)
                """

                ovalX = xscaler.convertFromScale(valX)
                ovalY = yscaler.convertFromScale(valY)

                # The training/validation sets combined:
                ocX = np.vstack((oX,ovalX))
                ocY = np.vstack((oY,ovalY))
            else:
                # If forcing is not done, rename
                ocX = oX
                ocY = oY

            #print('shape(ocX):',np.shape(ocX))
            #print('shape(ocY):',np.shape(ocY))


            # Separate ocX and ocY into nF separate folds
            nF = int(np.round(1/validationSetP))
            #print('Total N:',len(ocY))
            if deterministicDobSCV:
                raise NotImplementedError
                #(Folds, rInds) = FoldSortMin(ocY[:],nF)
                (valInds,valX), (trnInds,trnX) = MinSorter2(ocY[:],nF,sel=-2)

            else:
                (Folds, rInds) = dobscv_SC(ocY[:],nF,deterministic=False)

                valInds = rInds[0]
                trnInds = rInds[1]
                #raise NotImplementedError('Tarkista, että onko val/tre oikein päin')
                for i in range(2,len(rInds)):
                    trnInds = np.hstack((trnInds,rInds[i]))

            np.save('Data/Saved_datasets/Splits/{0:s}{3:s}_valInds_o{2:d}_itrn{1:d}.npy'.format(
                dkey,itrn,outputSel,titlevar),valInds)
            np.save('Data/Saved_datasets/Splits/{0:s}{3:s}_trnInds_o{2:d}_itrn{1:d}.npy'.format(
                dkey,itrn,outputSel,titlevar),trnInds)



def getArtificialDatasets():
    return {
        #'C00-2':complexity00,
        #'C01-2':complexity01,
        #'C02-2':complexity02,
        #'C03-2':complexity03,
        #'C04-10':complexity04,
        #'C05-5':complexity05,
        #'C06-6':complexity06,
        #'C07-6':complexity07,
        #'C08-6':complexity08,
        #'C09-8':complexity09,
        #'C10-15':complexity10,
        #'C11-10':complexity11,
        #'C12-10':complexity12,
        #'C13-10':complexity13,
        #'C14-10':complexity14,
        #'The100':The100,
        }


def savePreprocessedDatasets_OnlyCorrectFeatures(dskeys):
    """
    """
    datasets = getArtificialDatasets()

    dsf = Path('Experiments/Saved_datasets')
    parf = Path(dsf,'Params')

    def fpth(pth,add):
        return Path(pth,add).as_posix()

    dks = [
        ('C00-2_1000',-1,0),
        ('C01-2_1000',-1,0),
        ('C02-2_1000',-1,0),
        ('C03-2_2000',-1,0),
        ('C04-10_2000',-1,0),
        ('C05-5_2000',-1,0),
        ('C06-6_2000',-1,0),
        ('C07-6_2500',-1,0),
        ('C08-6_3000',-1,0),
        ('C09-8_3000',-1,0),
        ('C10-15_3000',-1,0),
        ('C11-10_5000',-1,0),
        ('C12-10_5000',-1,0),
        ('C13-10_3000',-1,0),
        ('C14-10_3000',-1,0),
        #('Madelon',-1,1),
        #('Gisette',-1,1),
        ]

    #Nls = [200,500,1000,1500,2000,2500,3000]
    #for N in Nls:
    for dskeyT in dks:
        dsName = dskeyT[0]
        (dskey, N) = dsName.split('_')
        N = int(N)
        df = datasets[dskey]
        res = df(N=N,unModOrig=True,onlyCorrect=True)
        X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask = res


        #if dskey=='SCM1D':
        #    Y = Y[:,0]
        #    Y = Y.reshape((-1,1))
        #    nY = Yscaler.Scaling(Y)

        add = 'ocF'
        np.save(fpth(dsf,'{0:s}_{1:d}_{2:s}_origX.npy').format(dskey,N,add),X)
        np.save(fpth(dsf,'{0:s}_{1:d}_{2:s}_origY.npy').format(dskey,N,add),Y)
        np.save(fpth(dsf,'{0:s}_{1:d}_{2:s}_normX.npy').format(dskey,N,add),nX)
        np.save(fpth(dsf,'{0:s}_{1:d}_{2:s}_normY.npy').format(dskey,N,add),nY)
        (xmin,xmax) = Xscaler.getScalingParams()
        (ymin,ymax) = Yscaler.getScalingParams()
        np.save(fpth(parf,'{0:s}_{1:d}_{2:s}_xmin.npy').format(dskey,N,add),xmin)
        np.save(fpth(parf,'{0:s}_{1:d}_{2:s}_xmax.npy').format(dskey,N,add),xmax)
        np.save(fpth(parf,'{0:s}_{1:d}_{2:s}_ymin.npy').format(dskey,N,add),ymin)
        np.save(fpth(parf,'{0:s}_{1:d}_{2:s}_ymax.npy').format(dskey,N,add),ymax)
        np.save(fpth(parf,'{0:s}_{1:d}_{2:s}_xconstmask.npy').format(dskey,N,add),
                XconstFiltMask)
        np.save(fpth(parf,'{0:s}_{1:d}_{2:s}_yconstmask.npy').format(dskey,N,add),
                YconstFiltMask)

        print(dskey)
        print(np.shape(nX))
        print(np.shape(nY))


if __name__ == "__main__":
    
    #savePreprocessedDatasets(getArtificialDatasets())
    #savePreprocessedDatasets_OnlyCorrectFeatures(getArtificialDatasets())
    CreateSyntheticDatasets(N=1000,noise=True)


    print('\nEnd of Line')
