#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Extreme Minimal Learning Machine (EMLM) implementation for regression

@author: Joonas Hämäläinen
@author: Joakim Linja
"""
import numpy as np
import warnings
from mlm import rps
from pairwiseDistanceCalculation import pdc
from ML_functions import solve_Axb, blockMatMul


def train(X, Y, beta=np.sqrt(np.finfo(float).eps),
          K_rel=100, rs_method="random", method="svd", 
          blockN:int =-1):
    """
    Trains EMLM model.

    Parameters
    ----------

    *X* : numpy.ndarray
        Input data matrix, rows are observations.

    *Y* : numpy.ndarray
        Output data matrix, rows are observations.

    *beta* : float
        small constant to assure unique solvability of the weights.

    *K_rel* : int, optional
        Relative #observations to be seleted as reference points.

        Default: 100

    *rs_method* : str, optional
        "random" uses random sampling to select reference points.

        "RS-maximin" uses RS-maximin to select reference points.

        Default: "random"
    
    *method* : str, optional
        Available options: 
            "lstsq"
            "lstsqF"
            "slstsq"
            "svd"
            "rSVD"
        
        Default: "svd"
        
        Implementations found in ML_functions.solve_Axb
        
    *blockN* : int, optional
        blockN defines the number of block matrices the computation is 
        split into. Disable with -1. 
        
        Default: -1
        

    Returns
    -------

    *W* : numpy.ndarray
        Weights of the trained EMLM model.

    *R* : numpy.ndarray
        Input data reference points

    """

    # X and Y should be 2D ndarrays, e.g., shape should be (N,1)
    # not (N,)
    if len(X.shape) == 1:
        warnings.warn("Reshaping input data matrix...")
        X = X.reshape(-1,1)
    if len(Y.shape) == 1:
        warnings.warn("Reshaping output data matrix...")
        Y = Y.reshape(-1,1)

    # Check for constant variables
    X_max = np.max(X,axis=0)
    X_min = np.min(X,axis=0)
    is_const_var = X_max - X_min == 0
    try:
        if any(is_const_var):
            warnings.warn("Input data matrix contains constant columns!" \
                          "Results may be inaccurate!")
    except ValueError as e:
        print(e)
        if is_const_var.any():
            warnings.warn("Input data matrix contains constant columns!" \
                          "Results may be inaccurate!")
    if K_rel==100:
        print('All points used as reference points.')
    else:
        print("Selecting reference points with " + rs_method + " method...")
    R,_ = rps(X,Y,K_rel,rs_method)

    W = _train(X,Y,R,beta,method=method,blockN=blockN)

    return W,R


def trainR(X, Y, R, beta=np.sqrt(np.finfo(float).eps),
           returnH=False, method='svd', blockN:int =-1):
    """
    Trains EMLM with selected sets of reference points.

    Parameters
    ----------

    *X* : numpy.ndarray
        Input data matrix, rows are observations.

    *Y* : numpy.ndarray
        Output data matrix, rows are observations.
        
    *R* : numpy.ndarray
        Input reference point matrix, rows are observations.

    *beta* : float
        small constant to assure unique solvability of the weights.

    *returnH* : boolean, optional
        Determines whether *H* matrix is returned along with *W*
    
        Default: False
        
    *method* : str, optional
        Available options: 
            "lstsq"
            "lstsqF"
            "slstsq"
            "svd"
            "rSVD"
        
        Default: "svd"
        
        Implementations found in ML_functions.solve_Axb
        
    *blockN* : int, optional
        blockN defines the number of block matrices the computation is 
        split into. Disable with -1. 
        
        Default: -1


    Returns
    -------

    *W* : numpy.ndarray
        Weights of the trained EMLM model.

    If *returnH* is True, also return:
    
    *H* : numpy.ndarray
        Distances between input data and reference points

    """

    # X and Y should be 2D ndarrays, e.g., shape should be (N,1)
    # not (N,)
    if len(X.shape) == 1:
        warnings.warn("Reshaping input data matrix...")
        X = X.reshape(-1,1)
    if len(Y.shape) == 1:
        warnings.warn("Reshaping output data matrix...")
        Y = Y.reshape(-1,1)

    # Check for constant variables
    X_max = np.max(X,axis=0)
    X_min = np.min(X,axis=0)
    is_const_var = X_max - X_min == 0
    if any(is_const_var):
        warnings.warn("Input data matrix contains constant columns!" \
                      "Results may be inaccurate!")

    if returnH:
        W, H = _train(X,Y,R,beta,method=method,returnH=returnH,blockN=blockN)
        return W,H
    else:
        W = _train(X,Y,R,beta,method=method,blockN=blockN)
        return W


def _train(X, Y, R, beta, returnH=False, method='svd', blockN:int =-1):
    """
    Trains EMLM with selected sets of reference points

    Parameters
    ----------

    *X* : numpy.ndarray
        Input data matrix, rows are observations.

    *Y* : numpy.ndarray
        Output data matrix, rows are observations.
        
    *R* : numpy.ndarray
        Input reference point matrix, rows are observations.

    *beta* : float
        small constant to assure unique solvability of the weights.

    *returnH* : boolean, optional
        Determines whether *H* matrix is returned along with *W*
    
        Default: False
        
    *method* : str, optional
        Available options: 
            "lstsq"
            "lstsqF"
            "slstsq"
            "svd"
            "rSVD"
        
        Default: "svd"
        
        Implementations found in ML_functions.solve_Axb
        
    *blockN* : int, optional
        blockN defines the number of block matrices the computation is 
        split into. Disable with -1. 
        
        Default: -1

    Returns
    -------

    *W* : numpy.ndarray
        Weights of the trained EMLM model.

    If *returnH* is True, also return:
    
    *H* : numpy.ndarray
        Distances between input data and reference points

    """

    # X,Y,R,T should be 2D ndarrays also for 1D cases,
    # e.g., shape should be (N,1) not (N,)
    if len(X.shape) == 1:
        warnings.warn("Reshaping input data matrix...")
        X = X.reshape(-1,1)
    if len(Y.shape) == 1:
        warnings.warn("Reshaping output data matrix...")
        Y = Y.reshape(-1,1)
    if len(R.shape) == 1:
        warnings.warn("Reshaping input data reference point matrix...")
        R = R.reshape(-1,1)

    # Compute the input space distance matrix
    H = pdc(R,X)

    # Solve output weights
    N = len(X); K = len(R)
    if blockN==-1:
        YH = Y.T @ H.T
        HHT = H @ H.T + (N*beta/K)*np.eye(K)
        WT = solve_Axb(HHT.T, YH.T, method=method)
        W = WT.T
    else:
        # Split @ operations into N block matrices when large matrices are assumed
        YH = blockMatMul(Y.T, H.T,blockN=blockN)
        HHT = blockMatMul(H,H.T,blockN=blockN) + (N*beta/K)*np.eye(K)
        WT = solve_Axb(HHT.T, YH.T, method=method,blockN=blockN)
        W = WT.T
    
    if returnH:
        return W, H
    else:
        return W


def pred(x, W, R, ay, by, blockN=-1, iterative=False):
    """
    Computes EMLM prediction y for given input x. Scaling for the Tommi's
    EMLM-matlab implementation.


    Parameters
    ----------
    *x* : numpy.ndarray
        Input vector

    *W* : numpy.ndarray
        Output weights.

    *R* : numpy.ndarray
        Input data reference points

    *ay* : numpy.ndarray
        Output scaling parameter

    *by* : numpy.ndarray
        Output scaling parameter
        
    *blockN* : int, optional
        blockN defines the number of block matrices the computation is 
        split into. Disable with -1. 
        
        Default: -1
        
    *iterative* : boolean, optional
        Determines whether the prediction is made iteratively one by one, or not.    
    
        Default: False

    Returns
    -------
    *y* : numpy.ndarray
        Predicted output.
    """

    
    if iterative:
        # Slow, but goes through the predicting iteratively
        (N,m) = np.shape(x)
        y = np.zeros((N,1))
        for i in range(N):
            ds = pdc(x[i,:], R)
            
            if blockN==-1:
                y[i] = ((W @ ds.T) + by)/ay
            else:
                y[i] = (blockMatMul(W,ds.T,blockN=blockN) + by)/ay
    else:
        ds = pdc(x, R)
        
        if blockN==-1:
            y = ((W @ ds.T) + by)/ay
        else:
            y = (blockMatMul(W,ds.T,blockN=blockN) + by)/ay

    return y.T


def pred2(x, W, R, blockN=-1, iterative=False):
    """
    Computes EMLM prediction y for given input x. No scaling of the output.


    Parameters
    ----------
    *x* : numpy.ndarray
        Input vector

    *W* : numpy.ndarray
        Output weights.

    *R* : numpy.ndarray
        Input data reference points
        
    *blockN* : int, optional
        blockN defines the number of block matrices the computation is 
        split into. Disable with -1. 
        
        Default: -1
        
    *iterative* : boolean, optional
        Determines whether the prediction is made iteratively one by one, or not.    
    
        Default: False


    Returns
    -------
    *y* : numpy.ndarray
        Predicted output.
    """
    return pred(x, W, R, 1, 0, blockN=blockN, iterative=iterative)


def pred_and_scale(x, W, R, ay, by, blockN=-1):
    """
    Computes EMLM prediction y for given input x. Scales the output.


    Parameters
    ----------
    *x* : numpy.ndarray
        Input vector

    *W* : numpy.ndarray
        Output weights.

    *R* : numpy.ndarray
        Input data reference points

    *ay* : numpy.ndarray
        Output scaling parameter

    *by* : numpy.ndarray
        Output scaling parameter
        
    *blockN* : int, optional
        blockN defines the number of block matrices the computation is 
        split into. Disable with -1. 
        
        Default: -1
        
    *iterative* : boolean, optional
        Determines whether the prediction is made iteratively one by one, or not.    
    
        Default: False

    Returns
    -------
    *y* : numpy.ndarray
        Predicted output.
    """
    #hx = euclidean_distances(R,x).reshape(-1,1)
    hx = pdc(R,x).reshape(-1,1)
    if blockN==-1:
        y = W @ hx
    else:
        y = blockMatMul(W,hx,blockN=blockN)
    y = (y - by)/ay
    return y