function [rX, rY, f2] = fid(X, Y, cutoff, plotToggle)
    % Feature Importance Detector, published in the paper: 
    % "Knowledge Discovery from Atomic Structures using Feature Importances" (arXiv:2303.09453). 
    % Description: 
    % Performs FID feature selection. 
    %
    % Inputs: 
    %           X - Index vector. 
    %           Y - Feature score vector. Assumed to be in descending
    %           order. 
    %           cutoff - The hyperparameter of FID. Range [0,1]. Determines the number
    %           of features to be removed. 
    %           plotToggle - Optional boolean input, default is false. 
    %
    % Ouputs:
    %           rX - Feature selected X
    %           rY - Feature selected Y
    %           f2 - Boolean vector, used to from rX and rY out of X and Y.
    %           

    if exist('plotToggle','var')
        plotT = plotToggle;
    else
        plotT = false;
    end
    
    fun = @(a,b,c,d,x) -a./(b*x+c)+d;

    x1 = X(1); y1 = Y(1);
    Y1 = zeros(size(Y));
    
    if plotT
        figure, plot(X,Y,'LineWidth',2);
        title('FID: Input')
    end
    
    Y1(2:end) = (Y(2:end,1)-y1)./(X(2:end,1)-x1);
    Y1(1) = Y1(2);
    Y1 = Y1/max(abs(Y1));
    
    if plotT
        figure, plot(X,Y1,'LineWidth',2);
        title('FID: Rolling slope')
    end
    
    fopt = fitoptions('Method','NonLinearLeastSquares','Lower',[0 -Inf -Inf -Inf],'Upper',[Inf Inf Inf Inf],'StartPoint',[1 10 10 0]);
    if plotT
        disp(fopt)
    end
    
    fitobj = fit(X,Y1,fun,fopt);
    
    Y2 = fitobj(X);
    
    if plotT
        figure, plot(X,Y2,'LineWidth',2);
        title('FID: Curve fit')
    end
    
    dY2 = Y2(end)-Y2(1);
    f2 = Y2 < Y2(end) - dY2*cutoff;
    
    rX = X(f2);
    rY = Y(f2);
    
    if plotT
        nf2 = ~f2;
        figure, plot(X,Y,'LineWidth',1);
        hold on;
        nrX = X(nf2);
        nrY = Y(nf2);
        plot(nrX,nrY,'LineWidth',2);
        title('FID: Result')
    end
end