#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Feature Importance Detector presented in  	
Knowledge Discovery from Atomic Structures using Feature Importances (arXiv:2303.09453)

@author: Joakim Linja
"""
import numpy as np
import scipy as sp

def fid(X,Y,cutoff = 0.01):
    """
    1. Bound rolling slope
    2. Curve fit a/(b*x+c)+d
    3. Remove features based on cutoff-defined percentage of total variance
    
    Parameters
    ----------

    X: numpy.ndarray
        Feature score index array in ascending numerical order.

    Y: numpy.ndarray
        Feature score array. Assumed to be sorted to descending order. 

    cutoff: float, optional, [0,1]
        Cutoff value to be used in the feature selection. 

        Default: 0.01


    Returns
    -------

    rX: numpy.matrix
        Feature selected X.

    rY: numpy.matrix
        Feature selected Y.

    f2: numpy.ndarray
        Boolean indexing array, the feature selection result.
        
    """
    def f(x,a,b,c,d):
        return -a/(b*x+c)+d
    
    x0 = X[0]; y0 = Y[0]
    Y1 = np.zeros(len(Y))
    for i in range(1,len(X)):
        Y1[i] = (Y[i]-y0)/(X[i]-x0)
    Y1[0] = Y1[1]
    Y1 = Y1/np.max(np.abs(Y1))
    
    
    popt, pcov = sp.optimize.curve_fit(f,X,Y1,p0=[1,10,10,0],
        bounds=([0,-np.inf,-np.inf,-np.inf],[np.inf,np.inf,np.inf,np.inf]))
    Y2 = f(X,*popt)
    
    dY2 = Y2[-1]-Y2[0]
    f2 = Y2 < Y2[-1] - dY2*cutoff
    
    rX = X[f2]
    rY = Y[f2]
    
    return rX, rY, f2
