#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A collection of feature ranking functions used with MLM/EMLM

@author: Joakim Linja
"""
import numpy as np
import scipy.stats as scs
from sklearn.feature_selection import mutual_info_regression, mutual_info_classif
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from mlm import rps
import sklearn_relief as sRelief
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.feature_selection import SelectFromModel
import skfeature_fisher_score as skffs


def _detectRegression(Y) -> bool:
    raise NotImplementedError('This does not work.')
    iY = Y.astype(np.int)
    return np.linalg.norm(Y-iY) > 1e-15


def spearmanR(X, Y, filterVar=0.05, returnValues=False, printT=True,
              regression=1, yscaler = None):
    """
    Pearson assumes normal distributions -> bad
    https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.stats.pearsonr.html
    Spearman does not care about distributions -> good
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.spearmanr.html
    Kendall is yet unsure
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kendalltau.html
    """
    (xi,xj) = np.shape(X)
    (yi,yj) = np.shape(Y)
    spearmanr = np.zeros((xj,2))
    if printT: print('Using SpearmanR')

    for i in range(xj):
        spearmanr[i,:] = scs.spearmanr(X[:,i],Y[:,0])

    spr = np.abs(spearmanr[:,0])
    return spr


def mutual_info(X, Y, printT=True, regression=1, yscaler = None):
    """
    """
    if regression:
        if printT: print('Using regression mutual information')
        return mutual_info_regression(X,Y[:,0])
    else:
        if printT: print('Using classification mutual information')
        if yscaler is None:
            raise NotImplementedError('Conversion to integers missing')
        iY = yscaler.convertFromScale(Y[:,0]).reshape(-1,1)
        iY2 = iY.astype(np.int)
        c = np.linalg.norm(iY-iY2) > 1e-15
        if c:
            iY = iY.astype(np.int)
        return mutual_info_classif(X,iY.reshape(-1,))


def RreliefF(X, Y, printT=True, regression=1, yscaler = None):
    """
    """
    if regression:
        if printT: print('Using RreliefF')
        SR = sRelief.RReliefF()
        SR.fit(X, Y[:,0])
        res = SR.w_
        del SR
        return np.abs(res)
    else:
        if printT: print('Using ReliefF classif')
        if yscaler is None:
            raise NotImplementedError('Conversion to integers missing')
        SR = sRelief.ReliefF()
        iY = yscaler.convertFromScale(Y[:,0])
        iY2 = iY.astype(np.int)
        c = np.linalg.norm(iY-iY2) > 1e-15
        if c:
            iY = iY.astype(np.int)
        SR.fit(X, iY)
        res = SR.w_
        del SR
        return np.abs(res)
        raise NotImplementedError


def decisionTree(X, Y, printT=True, regression=1, yscaler = None):
    """
    https://heartbeat.fritz.ai/hands-on-with-feature-selection-techniques-filter-methods-f248e0436ce5
    """
    (xi,xj) = np.shape(X)

    if regression:
        if printT: print('Using regression tree')
        yY = Y
    else:
        if printT: print('Using classification tree')
        if yscaler is None:
            raise NotImplementedError('Conversion to integers missing')
        iY = yscaler.convertFromScale(Y)
        iY2 = iY.astype(np.int)
        c = np.linalg.norm(iY-iY2) > 1e-15
        if c:
            iY = iY.astype(np.int)
        yY = iY

    (_,_,I) = rps(X,yY,85,'RS-maximin',returnI=True)
    trainX = X[I,:]
    trainY = yY[I,:]
    nr = np.full((xi,),True)
    nr[I] = False

    scores = -1*np.ones((xj,))
    
    if regression:
        clf = DecisionTreeRegressor()
    else:
        clf = DecisionTreeClassifier()
        raise NotImplementedError 
    clf.fit(trainX, trainY)
    scores = clf.feature_importances_
   
    return np.array(scores)


def RandomForestSelector(X, Y, N=100, printT=True,regression=1, yscaler = None):
    return RandomForest(X, Y, N=N, autoselector=True, printT=printT,
                        regression=regression, yscaler = yscaler)


def RandomForestRanker(X, Y, N=100, printT=True,regression=1, yscaler = None):
    return RandomForest(X ,Y, N=N, autoselector=False, printT=printT,
                        regression=regression, yscaler = yscaler)


def RandomForest(X, Y, N=100, autoselector=False, printT=True,
                 regression=1, yscaler = None):
    """
    Assumption: regression if not stated otherwise
    """

    if regression==1:
        if printT: print('Using regression forest')
        yY = Y
        rfm = RandomForestRegressor(n_estimators = N,
                            n_jobs=-1, oob_score=True)
    else:
        if printT: print('Using classification forest')
        if yscaler is None:
            raise NotImplementedError('Conversion to integers missing')
            
        iY = yscaler.convertFromScale(Y)
        iY2 = iY.astype(np.int)
        c = np.linalg.norm(iY-iY2) > 1e-15
        if c:
            iY = iY.astype(np.int)
        yY = iY
        rfm = RandomForestClassifier(n_estimators = N,
                            n_jobs=-1, oob_score=True)

    if autoselector:
        sel = SelectFromModel(rfm)
        sel.fit(X, yY.reshape(-1,))
        oob = sel.estimator_.oob_score_
        res = sel.get_support()

    else:
        rfm.fit(X, yY.reshape(-1,))
        oob = rfm.oob_score_
        res = rfm.feature_importances_

    return res, oob


def fisher_score(X, Y, printT=True, regression=1, yscaler = None):
    if printT: print("Using Fisher's score")
    return skffs.feature_ranking(skffs.fisher_score(X, Y[:,0]))


def MeanAbsoluteDifference(X, Y, printT=True, regression=1, yscaler = None):
    """
    https://www.analyticsvidhya.com/blog/2020/10/feature-selection-techniques-in-machine-learning/
    21.10.2021
    """
    if printT: print("Using Mean Absolute Difference")
    mean_abs_diff = np.sum(np.abs(X-np.mean(X,axis=0)), axis=0)/X.shape[0]
    return mean_abs_diff


def getFilters():
    return {
        'SpearmanR':spearmanR,
        'mutual_info':mutual_info,
        'RreliefF':RreliefF,
        'DecisionTree':decisionTree,
        'RandomForestSelector':RandomForestSelector,
        'RandomForestRanker':RandomForestRanker,
        'fisher-score':fisher_score,
        'MAD':MeanAbsoluteDifference,
        }

