#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Collection of functions made to behave as if they were in Matlab, due to 
many of the MLM/EMLM codes originating from Matlab codes.

@author: Joakim Linja
"""

from typing import Callable
import numpy as np
from pairwiseDistanceCalculation import pdc
from scipy.stats import kruskal as kruskalwallis


def categoricalKruskalWallisHtest(X: np.ndarray, lblvec: np.ndarray):
    """
    Kruskal-Wallis H-test that groups *X* according to the contents of
    *lblvec* and then uses the grouped *X* to compute H-test score
    which is also the return value.

    Source: https://stackoverflow.com/questions/51632900/pandas-apply-kruskal-wallis-to-numeric-columns
    https://stackoverflow.com/questions/35276217/use-groups-in-scipy-stats-kruskal-similar-to-r-cran-kruskal-test
    
    Parameters
    ----------
    *X* : numpy.ndarray
        Input data matrix
        
    *lblvec* : numpy.ndarray
        A vector of labels to group *X* by.
    
    """
    def groupByLabels(x: np.ndarray, lbls: np.ndarray):
        """
        Groups *x* by *lbls*.
        """
        assert len(x) == len(lbls)
        slbls = set(lbls)
        N = len(slbls)
        #print(N,'unique labels in grouping')
        grouped = [[] for a in slbls]
        llbls = list(slbls)

        for i in range(N):
            lbl = llbls[i]
            mask = lbls==lbl
            grouped[i] = x[mask]

        return grouped

    grouped = groupByLabels(X, lblvec)
    [res, _] = kruskalwallis(*grouped)
    return res


def hist(X:np.ndarray, binCount):
    """
    Returns bin counts and bin centers
    """
    (counts, edges) = np.histogram(X, bins=binCount)
    centers = np.zeros((binCount,))
    for i in range(binCount):
        centers[i] = (edges[i+1]-edges[i])/2 +edges[i]
    assert len(counts) == len(centers)
    return (counts, centers)


def bsxfun(bo: Callable, a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """
    Elementwise binary operation between a and b
    """
    TE = TypeError(
            'Inputs *a* and *b* must either be 1D vectors or of same shape')
    assert type(a)==np.ndarray
    assert type(b)==np.ndarray
    shpa = np.shape(a)
    shpb = np.shape(b)

    if len(shpa)==1 or len(shpb)==1:
        return bo(a,b)

    (ai,aj) = shpa
    (bi,bj) = shpb
    if ai!=1 or aj!=1 or bi!=1 or bj!=1:
        if ai==bi and aj==bj:
            return bo(a,b)
        else:
            raise TE
    elif (ai==1 or aj==1) and (bi==1 or bj==1):
        return bo(a,b)
    else:
        raise TE


def weightedEuclideanDistance(A, B, W):
    """
    $$d(A,B) = \sqrt{ \sum_i w_i (A_i - B_i)^2 }$$
    """
    nma = np.shape(A)
    nmb = np.shape(B)
    nmw = np.shape(W)
    try:
        if len(nma)==1:
            assert nma == nmb
            assert nma == nmw
        elif len(nma)==2:
            assert nma[1]==nmb[1]
            assert nma[1]==nmw[1]
        else:
            raise NotImplementedError
        res = np.zeros((nma[0],nmb[0]))
        WW = W[0,:]*W[0,:]
        for i in range(nma[0]):
            for j in range(nmb[0]):
                res[i,j] = np.sqrt(np.sum(WW*(A[i,:]-B[j,:])**2))
        return res
    except AssertionError:
        print('nma:',nma)
        print('nmb:',nmb)
        print('nmw:',nmw)
        raise AssertionError


def pdist2(a: np.ndarray, b: np.ndarray,
           sort: None or str=None, n: int=-1):
    """
    # [~,I] = pdist2(CData,e0,'euclidean','Smallest',nfolds);
    # pdc that returns nfolds-number of smallest distances
    (_, I) = pdist2(CData, e0, sort='Smallest', n=nfolds)
    """

    (ai,aj) = np.shape(a)
    (bi,bj) = np.shape(b)
    assert aj==bj

    dists = pdc(a,b)

    if sort is None:
        return dists

    distsShp = np.shape(dists)
    (di,dj) = distsShp

    if sort in 'Smallest':
        inds = np.zeros(distsShp,dtype=np.int)
        for j in range(dj):
            inds[:,j] = np.argsort(dists[:,j])
            dists[:,j] = dists[inds[:,j],j]

        if n!=-1:
            return (dists[:n,:], inds[:n,:])
        return (dists, inds)
    else:
        raise NotImplementedError


    raise NotImplementedError


if __name__ == "__main__":



    print('\nEnd of Line')
