#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Scaling class prototype and MinMax implementation for use with MLM & EMLM


@author: Joakim Linja
@author: Joonas Hämäläinen
"""

import numpy as np


class Scaler:
    def __init__(self):
        raise NotImplementedError

    def setScalingParams(self, inpt):
        raise NotImplementedError

    def getScalingParams(self):
        raise NotImplementedError

    def convertToScale(self, x):
        raise NotImplementedError

    def convertFromScale(self, x):
        raise NotImplementedError

    def Scaling(self, X):
        raise NotImplementedError


class MinMax(Scaler):
    """
    MinMax scaling class
    
    ...
    
    Attributes
    ----------
    
    
    Methods
    -------
    setScalingParams(inpt)
        Manually set minimum and maximum values of the input data according to 
        tuple *inpt*.
    
    getScalingParams()
        Get minimum and maximum values of the input data.
        
    convertToScale()
        Converts *X* from original scale to MinMax scale if scaling parameters 
        have been set. 
    
    convertFromScale()
        Converts *X* from MinMax scale to original scale.
        
    Scaling()
        Converts *X* into MinMax-scale and returns it. Automatically sets 
        scaling parameters. 
    
    """
    
    def __init__(self, interval=[0,1]):
        """
        Initialize MinMax scaler to [0,1] interval by default or to another 
        set by user. Defaults minimum and maximum feature arrays to 0. 
        Function *Scaling* automatically sets them based on the given data. 
        
        Parameters
        ----------
        *interval* : tuple, optional. Default value is [0,1]. 
            Sets the interval for the MinMax scaler class. 
        """
        
        self._minX = 0
        self._maxX = 0
        self._Imin = interval[0]
        self._Imax = interval[1]


    def setScalingParams(self, inpt: tuple):
        """
        Manually sets scaling parameters.
        
        Parameters
        ----------
        *inpt* : (numpy.ndarray, numpy.ndarray) 
            Tuple containing an array of the minimum value of each feature and 
            the maximum value of each feature. 
        """
        
        (self._minX, self._maxX) = inpt


    def getScalingParams(self):
        """
        Retrieve current stored scaling parameters. 
        
        Returns
        -------
        
        (array of feature minimums, array of feature maximums)
        """
        
        return (self._minX, self._maxX)


    def _handleInput(self, x: np.ndarray, minX: np.ndarray, maxX: np.ndarray,
                     mask: np.ndarray):
        """
        If scaling parameters are not given, they are taken from memory,
        which was updated when Scaling was last called. 
        This function is not called manually. 
        
        Parameters
        ----------
        *x* : numpy.ndarray
            Input data matrix, rows are observations.
        
        *minX* : numpy.ndarray
            Array of feature minimums. Previously stored one is used in the case 
            *minX* is None.
        
        *maxX* : numpy.ndarray
            Array of feature maximums. Previously stored one is used in the case 
            *maxX* is None.
        
        *mask* : boolean numpy.ndarray
            Constant feature filter array. Is used to remove constants if 
            *mask* is provided to functions which call this function. 
        
        Returns
        -------
            (array of feature minimums, array of feature maximums)
        
        Raises
        ------
        
        ZeroDivisionError
            In the case set scaling parameters would cause arithmetic errors. 
            Should not happen when scaling parameters are set automatically. 
        """
        
        if minX is None:
            miX = self._minX
        else:
            miX = minX
        if maxX is None:
            maX = self._maxX
        else:
            maX = maxX

        if mask is not None:
            if len(np.shape(x))>1:
                x = x[:,mask]
            miX = miX[mask]
            maX = maX[mask]
        self._check(miX,maX)
        return (miX, maX)


    def _check(self, minX, maxX, mask=None):
        """
        Checks if the given feature minimums and maximums would cause a 
        zero division error. 
        
        Parameters
        ----------
        
        *minX* : numpy.ndarray
            Array of feature minimums. 
        
        *maxX* : numpy.ndarray
            Array of feature maximums. 
        
        *mask* : boolean numpy.ndarray, optional
            Constant feature filter array. Is used to remove constants if provided. 
        
        Raises
        ------
        
        ZeroDivisionError
            In the case dividend would be zero
        """
        
        a = np.abs(maxX-minX)
        if mask is not None:
            a = a[mask]
        if np.shape(minX) == () and np.shape(maxX) == ():
            c = a <1e-16
        else:
            t = a<1e-16
            if len(np.shape(t))>1:
                t = t.any()
            else:
                t = any(t)
            c = t
        if c:
            raise ZeroDivisionError('Dividend would be {0}, {1}'.format(
                    a,np.shape(a)))


    def convertToScale(self, x: np.ndarray,
                       minX: np.ndarray=None, maxX: np.ndarray=None,
                       mask: np.ndarray=None):
        """
        Converts *x* to MinMax scale
        
        Parameters
        ----------
        *x* : numpy.ndarray
            Input data matrix to be converted, rows are observations. 
        
        *minX* : numpy.ndarray, optional
            Array of feature minimums. Previously stored one is used in the case 
            *minX* is not provided.
        
        *maxX* : numpy.ndarray, optional
            Array of feature maximums. Previously stored one is used in the case 
            *maxX* is not provided.
        
        *mask* : boolean numpy.ndarray, optional
            Constant feature filter array. Is used to remove constants if provided. 
        
        
        Returns
        -------
        
        *x* with features converted to set MinMax interval. 
        """
        
        (miX, maX) = self._handleInput(x, minX, maxX, mask)

        k1 = (self._Imax-self._Imin)/(self._maxX-self._minX)
        k2 = self._Imin-k1*self._minX
        return x*k1+k2


    def convertFromScale(self, x: np.ndarray,
                       minX: np.ndarray=None, maxX: np.ndarray=None,
                       mask: np.ndarray=None):
        """
        Converts *x* from MinMax scale to original
        
        Parameters
        ----------
        *x* : numpy.ndarray
            Input data matrix to be converted, rows are observations.
        
        *minX* : numpy.ndarray, optional
            Array of feature minimums. Previously stored one is used in the case 
            *minX* is not provided.
        
        *maxX* : numpy.ndarray, optional
            Array of feature maximums. Previously stored one is used in the case 
            *maxX* is not provided.
        
        *mask* : boolean numpy.ndarray, optional
            Constant feature filter array. Is used to remove constants if provided. 
        
        
        Returns
        -------
        
        *x* with features converted from MinMax interval to original. 
        """
        
        (miX, maX) = self._handleInput(x, minX, maxX, mask)
        
        k1 = (self._Imax-self._Imin)/(self._maxX-self._minX)
        k2 = self._Imin-k1*self._minX
        return (x-k2)/k1


    def Scaling(self, X, mask=None):
        """
        Converts *X* into MinMax-scale and returns converted with the
        minimum and maximum of *X*. Sets scaling parameters automatically. 
        
        Parameters
        ----------
        *X* : numpy.ndarray
            Dataset to be scaled to previously set MinMax interval. 
        
        *mask* : numpy.ndarray, optional
            Constant feature filter array. Is used to remove constants if provided. 
        """
        
        minX = np.min(X,axis=0)
        maxX = np.max(X,axis=0)
        self._check(minX, maxX, mask=mask)
        self._minX = minX
        self._maxX = maxX
        nX = self.convertToScale(X, mask=mask)
        return nX


if __name__ == "__main__":
    MM = MinMax()
    X = np.random.rand(5,5)*100


    nX = MM.Scaling(X)
    print('Scaling params')
    print(MM.getScalingParams())

    print('X')
    print(X)
    print('nX')
    print(nX)
    print('returned X')
    print(MM.convertFromScale(nX))

    X2 = np.random.rand(1,5)*100
    nX2 = MM.convertToScale(X2)
    print('X2')
    print(X2)
    print('nX2')
    print(nX2)
    print('returned X2')
    print(MM.convertFromScale(nX2))

    MM2 = MinMax()
    X3 = np.random.rand(10,4)
    nX3 = MM.Scaling(X3)
    print('X3')
    print(X3)
    print('nX3')
    print(nX3)
    print('returned X3')
    print(MM.convertFromScale(nX3))


    print('\nEnd of Line')
